# Copyright (c) 2017-2021 Scott Coughlin
#               2021 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Extend :mod:`astropy.table` with the `GravitySpyTable`."""

from __future__ import annotations

import logging
import re
from concurrent.futures import ThreadPoolExecutor
from json import JSONDecodeError
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import urlencode

import requests

from .table import EventTable

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Self

    from astropy.table import Row

__author__ = "Scott Coughlin <scott.coughlin@ligo.org>"
__all__ = ["GravitySpyTable"]

logger = logging.getLogger(__name__)

JSON_CONTENT_TYPE = "application/json"

# default gravity spy remote host
DEFAULT_HOST = "https://gravityspytools.ciera.northwestern.edu"

# API paths
SEARCH_PATH = "/search"
SIMILARITY_SEARCH_PATH = "/similarity_search_restful_API"

# ordered list of image durations supported by gravityspy.
# NOTE: the index of each element here is the map between
#       the 'urlX' column in the GravitySpyTable and the
#       duration of the relevant image, for some reason.
DURATIONS = (
    0.5,
    1.0,
    2.0,
    4.0,
)

# list of ERAs supported by similarity search API
ERA = {
    "ALL": (1126400000, 1584057618),
    "O1": (1126400000, 1137250000),
    "ER10": (1161907217, 1164499217),
    "O2a": (1164499217, 1219276818),
    "ER13": (1228838418, 1229176818),
    "ER14": (1235750418, 1238112018),
    "O3": (1238166018, 1272326418),
    "O3a": (1238166018, 1254009618),
    "O3b": (1256655642, 1272326418),
}

# regex to filter columns containing URLs
URL_COLUMN = re.compile(r"\Aurl[0-9]+\Z")


def _image_download_target(
    row: Row,
    outdir: Path,
    *,
    training_set: bool = False,
    labelled_samples: bool = False,
) -> str:
    """Construct where to download a GravitySpy image.

    This returns a `str.format`-style template that just needs
    the duration to form a complete target path.
    """
    ifo = row["ifo"]
    id_ = row["gravityspy_id"]
    label = row["ml_label"] if training_set else ""
    stype = row["sample_type"] if labelled_samples else ""
    return str(
        outdir
        / label
        / stype
        / f"{ifo}_{id_}_spectrogram_{{}}.png",
    )


def _download_image(
    bundle: tuple[str, Path],
    timeout: float | None = 30.0,
    **kwargs,
) -> None:
    url, target = bundle
    logger.debug("Downloading %s to %s", url, target)
    resp = requests.get(
        url,
        timeout=timeout,
        **kwargs,
    )
    resp.raise_for_status()
    target.write_bytes(resp.content)
    logger.debug("Wrote %s", target)


class GravitySpyTable(EventTable):
    """A container for a table of Gravity Spy Events.

    As well as Events from the O1 Glitch Classification Paper which includes:

    - PCAT
    - PC-LIB
    - WDF
    - WDNN
    - Karoo GP

    See Also
    --------
    astropy.table.Table
        for details on parameters for creating a `GravitySpyTable`
    """

    # -- i/o -------------------------

    def download(
        self,
        download_path: str | Path = "download",
        parallel: int = 1,
        download_durs: Sequence[float] = DURATIONS,
        *,
        training_set: bool = False,
        labelled_samples: bool = False,
    ) -> list[str]:
        """Download image files associated with entries in a `GravitySpyTable`.

        Parameters
        ----------
        parallel : `int`, optional
            Number of parallel threads to use for parallel file reading.

        download_path : `str` optional
            Target directory for downloaded images.

        download_durs : `list` optional
            List of durations to download, must be a subset of
            ``[0.5, 1.0, 2.0, 4.0]``.
            Default is to download all the available GSpy durations.

        training_set : `bool`, optional
            If `True` download training set data.

        labelled_samples : `bool`, optional
            If `True` download only labelled samples.
        """
        download_path = Path(download_path)

        # work out which columns contain URLs
        url_columns = list(filter(URL_COLUMN.match, self.colnames))

        # construct a download target location for each requested image
        urls: dict[str, Path] = {}
        for row in self:
            if row["url1"] in {None, "", "?"}:  # skip rows without URLs
                continue

            # construct the abstract download target
            _download_target = _image_download_target(
                row,
                download_path,
                training_set=training_set,
                labelled_samples=labelled_samples,
            )

            # and map the remote URL to a concrete target
            # for each duration requested
            for col in url_columns:
                idx = int(col[3:])  # urlXXXX number
                duration = DURATIONS[idx-1]
                if duration in download_durs:  # image requested
                    urls[row[col]] = Path(_download_target.format(duration))

        # create all of the directories needed up front
        for path in urls.values():
            path.parent.mkdir(exist_ok=True, parents=True)

        # download the images
        with ThreadPoolExecutor(max_workers=parallel) as executor:
            executor.map(
                _download_image,
                list(urls.items()),
            )

        # and return the list of new files we have
        return sorted(map(str, urls.values()))

    @classmethod
    def search(
        cls,
        gravityspy_id: str,
        howmany: int = 10,
        era: str = "ALL",
        ifos: tuple[str, ...] | str = ("H1", "L1"),
        database: str = "similarity_index_o3",
        host: str = DEFAULT_HOST,
        timeout: float | None = 30.0,
        **kwargs,
    ) -> Self:
        """Perform a GravitySpy 'Similarity Search' for the given ID.

        Parameters
        ----------
        gravityspy_id : `str`
            The unique 10 character hash that identifies a Gravity Spy image.

        howmany : `int`, optional
            The number of similar images you would like.

        era : `str`, optional
            Which era to search, see online for more information.

        ifos : `tuple`, str`, optional
            The list of interferometers to include in the search.

        database : `str`, optional
            The database to query.

        host : `str`, optional
            The URL (scheme and FQDN) of the Gravity Spy host to query

        timeout : `float`, optional
            How long to wait for the server to respond before giving up.

        **kwargs
            Other kwargs are passed to `requests.get` to perform
            the actual remote communication.

        Returns
        -------
        table : `GravitySpyTable`
            A `GravitySpyTable` containing similar events based on
            an evaluation of the Euclidean distance of the input image
            to all other images in some Feature Space.

        Notes
        -----
        For an online version, and documentation of the search, see

        https://gravityspytools.ciera.northwestern.edu/search/
        """
        if isinstance(ifos, str):  # 'H1L1' -> ['H1', 'L1']
            ifo_list = [ifos[i:i+2] for i in range(0, len(ifos), 2)]
        else:
            ifo_list = list(ifos)

        query = urlencode({
            "howmany": howmany,
            "imageid": gravityspy_id,
            "era": "event_time BETWEEN {} AND {}".format(*ERA[era]),
            "ifo": ", ".join(map(repr, ifo_list)),
            "database": database,
        })
        url = host + SEARCH_PATH + SIMILARITY_SEARCH_PATH + "/?" + query

        logger.debug("Executing GravitySpy similarity search: %s", url)
        response = requests.get(url, timeout=timeout, **kwargs)
        logger.debug("Response: %s", response)
        response.raise_for_status()  # check the response
        try:
            return GravitySpyTable(response.json())
        except JSONDecodeError:  # that didn't work
            # if the response was actually HTML, something terrible happened
            if response.headers["Content-Type"] != JSON_CONTENT_TYPE:
                msg = (
                    f"response from {response.url} was '200 OK' but "
                    "the content is HTML, please check the request parameters"
                )
                raise requests.HTTPError(
                    msg,
                    response=response,
                ) from None
            raise



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.segments`."""

from __future__ import annotations

import re
import warnings
from contextlib import nullcontext
from http import HTTPStatus
from io import BytesIO
from typing import TYPE_CHECKING
from unittest import mock

import h5py
import pytest
from matplotlib import rc_context
from requests.exceptions import HTTPError
from requests.models import Response
from requests.status_codes import codes as status_codes

from ...plot import SegmentAxes
from ...segments import (
    DataQualityDict,
    DataQualityFlag,
    Segment,
    SegmentList,
)
from ...testing import utils
from ...testing.errors import pytest_skip_flaky_network

if TYPE_CHECKING:
    from typing import SupportsFloat

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- veto definer fixture ------------

VETO_DEFINER_FILE = """
<?xml version='1.0' encoding='utf-8'?>
<!DOCTYPE LIGO_LW SYSTEM "http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt">
<LIGO_LW>
    <Table Name="veto_definer:table">
        <Column Type="int_4s" Name="category"/>
        <Column Type="lstring" Name="comment"/>
        <Column Type="int_4s" Name="end_pad"/>
        <Column Type="int_8s" Name="process:process_id"/>
        <Column Type="lstring" Name="name"/>
        <Column Type="int_4s" Name="version"/>
        <Column Type="int_4s" Name="start_pad"/>
        <Column Type="int_4s" Name="start_time"/>
        <Column Type="lstring" Name="ifo"/>
        <Column Type="int_4s" Name="end_time"/>
        <Stream Delimiter="," Type="Local" Name="veto_definer:table">
            1,"Test flag 1",2,,"TEST-FLAG",1,-1,100,"X1",0,
            2,"Test flag 1",2,,"TEST-FLAG_2",1,1,100,"X1",200,
            2,"Test flag 1",2,,"TEST-FLAG_2",2,-2,200,"X1",0,
            2,"Test flag 1",2,,"TEST-FLAG_2",2,-2,100,"Y1",0
        </Stream>
    </Table>
</LIGO_LW>
""".strip()


@pytest.fixture
def veto_definer(tmp_path):
    """Write the ``VETO_DEFINER_FILE`` to a temporary file and return."""
    tmp = tmp_path / "veto-definer.xml"
    tmp.write_text(VETO_DEFINER_FILE)
    return tmp


def veto_def(ifo, name, version, **kwargs):
    """Create a `igwn_ligolw.lsctables.VetoDef`."""
    from igwn_ligolw.lsctables import VetoDef

    vdef = VetoDef()
    kwargs["ifo"] = ifo
    kwargs["name"] = name
    kwargs["version"] = version
    for key in VetoDef.__slots__:
        setattr(vdef, key, kwargs.get(key))
    return vdef

# -- XML without _ns columns ---------

LIGOLW_NO_NS = """
<?xml version="1.0"?>
<!DOCTYPE LIGO_LW SYSTEM "http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt">
<LIGO_LW>
  <Table Name="segment_definergroup:segment_definer:table">
    <Column Name="segment_def_id" Type="int_8s"/>
    <Column Name="ifos" Type="lstring"/>
    <Column Name="name" Type="lstring"/>
    <Column Name="version" Type="int_4s"/>
    <Column Name="comment" Type="lstring"/>
    <Stream Name="segment_definergroup:segment_definer:table" Type="Local" Delimiter=",">
      0,"X1","TEST_FLAG",1,"Test flag",
    </Stream>
  </Table>
  <Table Name="segment_summarygroup:segment_summary:table">
    <Column Name="segment_sum_id" Type="int_8s"/>
    <Column Name="start_time" Type="int_4s"/>
    <Column Name="end_time" Type="int_4s"/>
    <Column Name="comment" Type="lstring"/>
    <Column Name="segment_definer:segment_def_id" Type="int_8s"/>
    <Stream Name="segment_summarygroup:segment_summary:table" Type="Local" Delimiter=",">
      0,1366644592,1366644608,"",0,
    </Stream>
  </Table>
  <Table Name="segmentgroup:segment:table">
    <Column Name="segment_id" Type="int_8s"/>
    <Column Name="start_time" Type="int_4s"/>
    <Column Name="end_time" Type="int_4s"/>
    <Column Name="segment_definer:segment_def_id" Type="int_8s"/>
    <Stream Name="segmentgroup:segment:table" Type="Local" Delimiter=",">
      0,1366644592,1366644593,0,
    </Stream>
  </Table>
</LIGO_LW>
""".strip()  # noqa: E501


@pytest.fixture
def ligolw_no_ns(tmp_path):
    """Write the LIGO_LW segment file without nanoseconds columns to a file."""
    tmp = tmp_path / "test.xml"
    tmp.write_text(LIGOLW_NO_NS)
    return tmp


# -- test data -----------------------

def _as_segmentlist(*segments: tuple[SupportsFloat, SupportsFloat]) -> SegmentList:
    """Return ``segments`` as a `SegmentList`."""
    return SegmentList([Segment(a, b) for a, b in segments])


NAME = "X1:TEST-FLAG_NAME:0"

# simple list of 'known' segments
KNOWN = _as_segmentlist(
    (0, 3), (6, 7))

# simple  list of 'active' segments
ACTIVE = _as_segmentlist(
    (1, 2), (3, 4), (5, 7))

# intersection of above 'known' and 'active' segments
KNOWNACTIVE = _as_segmentlist(
    (1, 2), (6, 7))

# 'active' set contracted by 0.1 seconds
ACTIVE_CONTRACTED = _as_segmentlist(
    (1.1, 1.9), (3.1, 3.9), (5.1, 6.9))

# 'active' seg protracted by 0.1 seconts
ACTIVE_PROTRACTED = _as_segmentlist(
    (.9, 2.1), (2.9, 4.1), (4.9, 7.1))

# some more segments
KNOWN2 = _as_segmentlist(
    (100, 150))

ACTIVE2 = _as_segmentlist(
    (100, 101), (110, 120))

# padding
PADDING = (-0.5, 1)

# padded version of above 'known' segments
KNOWNPAD = _as_segmentlist(
    (-.5, 4), (5.5, 8))

# padded version of above 'active' segments
ACTIVEPAD = _as_segmentlist(
    (.5, 3), (2.5, 5), (4.5, 8))

# padded, coalesed version of above 'active' segments
ACTIVEPADC = _as_segmentlist(
    (.5, 4), (5.5, 8))


# -- query helpers -------------------

QUERY_FLAGS = ["X1:TEST-FLAG:1", "Y1:TEST-FLAG2:4"]

QUERY_RESULT = DataQualityDict()

QUERY_RESULT["X1:TEST-FLAG:1"] = DataQualityFlag(
    "X1:TEST-FLAG:1",
    known=[(0, 10)],
    active=[(0, 1), (1, 2), (3, 4), (6, 9)])

QUERY_RESULT["Y1:TEST-FLAG2:4"] = DataQualityFlag(
    "Y1:TEST-FLAG2:4",
    known=[(0, 5), (9, 10)],
    active=[])

QUERY_RESULTC = type(QUERY_RESULT)({x: y.copy().coalesce() for
                                    x, y in QUERY_RESULT.items()})


def mock_query_segments(flag, start, end, **kwargs):
    """Mock `dqsegdb2.query.query_segments`."""
    try:
        ifo, name, version = flag.split(":")
        version = int(version)
    except ValueError:
        ifo, name = flag.split(":", 1)
        version = None
    span = SegmentList([Segment(start, end)])
    reflag = re.compile(flag)
    try:
        actual = next(filter(reflag.match, QUERY_RESULT))
    except StopIteration:  # empty
        response = Response()
        response.status_code = status_codes.NOT_FOUND
        response.reason = HTTPStatus(response.status_code).phrase
        response.raw = BytesIO()
        response.raw.write(br"<h1>{response.statue_code} {response.reason}</h1>")
        msg = (
            f"{response.status_code} Client Error: "
            f"{response.reason} for url: something"
        )
        raise HTTPError(msg, response=response)
    return {
        "ifo": ifo,
        "name": name,
        "version": version,
        "known": list(map(tuple, QUERY_RESULT[actual].known & span)),
        "active": list(map(tuple, QUERY_RESULT[actual].active & span)),
        "query_information": {},
        "metadata": kwargs,
    }


# -- DataQualityFlag -----------------

class TestDataQualityFlag:
    """Test `gwpy.segments.DataQualityFlag."""

    TEST_CLASS = DataQualityFlag

    @classmethod
    def create(cls, name=NAME, known=KNOWN, active=ACTIVE, **kwargs):
        """Create a new `DataQualityFlag` for testing."""
        return cls.TEST_CLASS(name=name, known=known, active=active, **kwargs)

    @pytest.fixture
    @classmethod
    def flag(cls):
        """Return a new `DataQualityFlag` for testing."""
        return cls.create()

    @pytest.fixture
    @classmethod
    def empty(cls):
        """Return a new, empty `DataQualityFlag` for testing."""
        return cls.TEST_CLASS()

    # -- test properties -------------

    def test_name(self, empty, flag):
        """Test ``name`` property handling."""
        assert empty.name is None

        assert flag.name == NAME
        assert flag.ifo == NAME.split(":")[0]
        assert flag.tag == NAME.split(":")[1]
        assert flag.version == int(NAME.split(":")[2])

    def test_known(self, flag):
        """Test ``known`` property handling."""
        utils.assert_segmentlist_equal(flag.known, KNOWN)

    def test_known_empty(self, empty):
        """Test empty ``known`` property handling."""
        assert isinstance(empty.known, SegmentList)
        assert empty.known == []

    def test_known_cast(self):
        """Test that the ``known`` property casts to `SegmentList`."""
        new = self.TEST_CLASS()
        new.known = [(1, 2), (3, 4)]  # type: ignore[assignment]
        assert isinstance(new.known, SegmentList)
        assert all(isinstance(seg, Segment) for seg in new.known)

    def test_active(self, flag):
        """Test ``active`` property handling."""
        utils.assert_segmentlist_equal(flag.active, ACTIVE)

    def test_active_empty(self, empty):
        """Test empty ``active`` property handling."""
        assert isinstance(empty.active, SegmentList)
        assert empty.active == []

    def test_active_cast(self):
        """Test that the ``active`` property casts to `SegmentList`."""
        new = self.TEST_CLASS()
        new.active = [(1, 2), (3, 4)]  # type: ignore[assignment]
        assert isinstance(new.active, SegmentList)
        assert all(isinstance(seg, Segment) for seg in new.active)

    @pytest.mark.parametrize(("name", "texname"), [
        (None, None),
        ("test", "test"),
        ("test_one_two", r"test\_one\_two"),
    ])
    def test_texname(self, name, texname):
        """Test ``texname`` property handling."""
        flag = self.TEST_CLASS(name=name)
        assert flag.texname == texname

    def test_extent(self):
        """Test ``extent`` property handling."""
        flag = self.TEST_CLASS(known=[(4, 6), (100, 101)])
        assert flag.extent == (4, 101)

    def test_extent_empty(self, empty):
        """Test ``extent`` property error handling."""
        with pytest.raises(
            ValueError,
            match="empty list",
        ):
            empty.extent  # noqa: B018

    @pytest.mark.parametrize(("active", "livetime"), [
        pytest.param([], 0, id="empty"),
        pytest.param([(1, 2), (3, 10)], 8, id="empty"),
    ])
    def test_livetime(self, active, livetime):
        """Test ``livetime`` property handling."""
        flag = self.TEST_CLASS(active=active)
        assert flag.livetime == livetime

    @pytest.mark.parametrize(("active", "known", "regular"), [
        pytest.param(
            [(0, 1), (3, 4)],
            [(0, 1), (2, 3), (3, 4)],
            True,
            id="regular",
        ),
        pytest.param(
            [(0, 1), (2, 3), (3, 4)],
            [(0, 1), (3, 4)],
            False,
            id="irregular",
        ),
    ])
    def test_regular(self, active, known, regular):
        """Test the ``regular`` property."""
        flag = self.TEST_CLASS(active=active, known=known)
        assert flag.regular is regular

    @pytest.mark.parametrize(("padding", "result"), [
        (None, (0, 0)),
        ((0, 0), (0, 0)),
        ([-1, 2], (-1, 2)),
    ])
    def test_padding(self, padding, result):
        """Test the ``padding`` property."""
        flag = self.TEST_CLASS(padding=padding)
        assert isinstance(flag.padding, tuple)
        assert flag.padding == result

        flag.padding = [-1, 2]  # type: ignore[assignment]
        assert isinstance(flag.padding, tuple)
        assert flag.padding == (-1, 2)

        flag.padding = None  # type: ignore[assignment]
        assert flag.padding == (0, 0)

    def test_padding_deleter(self):
        """Test ``del flag.padding``."""
        flag = self.TEST_CLASS(padding=(1, 2))
        assert flag.padding == (1, 2)
        del flag.padding
        assert flag.padding == (0, 0)

    # -- test methods ----------------

    @pytest.mark.parametrize(("flag", "name", "ifo", "tag", "version"), [
        (None, None, None, None, None),
        ("test", "test", None, None, None),
        ("L1:test", "L1:test", "L1", "test", None),
        ("L1:test:1", "L1:test:1", "L1", "test", 1),
    ])
    def test_parse_name(self, flag, name, ifo, tag, version):
        """Test name parsing."""
        flag = self.TEST_CLASS(flag)
        assert flag.name == name
        assert flag.ifo == ifo
        assert flag.tag == tag
        assert flag.version == version

    def test_plot(self, flag):
        """Test `DataQualityFlag.plot()`."""
        flag.label = "Test label"
        with rc_context(rc={"text.usetex": False}):
            plot = flag.plot(figsize=(6.4, 3.8))
            assert isinstance(plot.gca(), SegmentAxes)
            assert plot.gca().get_epoch() == flag.known[0][0]
            assert len(plot.gca().collections) == 2
            assert len(plot.gca().collections[1].get_paths()) == len(KNOWN)
            assert len(plot.gca().collections[0].get_paths()) == len(ACTIVE)
            assert plot.gca().collections[0].get_label() == flag.label

            plot.save(BytesIO(), format="png")
            plot.close()

    def test_plot_texname(self, flag):
        """Test `DataQualityFlag.plot()`."""
        flag.label = None
        with rc_context(rc={"text.usetex": True}):
            plot = flag.plot(figsize=(6.4, 3.8))
            assert plot.gca().collections[0].get_label() == flag.texname
            plot.close()

    def test_and(self):
        """Test ``a & b``."""
        a = self.TEST_CLASS(active=ACTIVE[:2], known=KNOWN)
        b = self.TEST_CLASS(active=ACTIVE[2:], known=KNOWN)
        c = a & b
        utils.assert_segmentlist_equal(c.active, a.active & b.active)
        utils.assert_segmentlist_equal(c.known, a.known & b.known)

    def test_sub(self):
        """Test ``a - b``."""
        a = self.TEST_CLASS(active=ACTIVE[:2], known=KNOWN)
        b = self.TEST_CLASS(active=ACTIVE[2:], known=KNOWN)
        c = a - b
        utils.assert_segmentlist_equal(c.known, a.known & b.known)
        utils.assert_segmentlist_equal(c.active, (a.active - b.active) & c.known)

    def test_or(self):
        """Test ``a | b``."""
        a = self.TEST_CLASS(active=ACTIVE[:2], known=KNOWN)
        b = self.TEST_CLASS(active=ACTIVE[2:], known=KNOWN)
        c = a | b
        utils.assert_segmentlist_equal(c.active, a.active | b.active)
        utils.assert_segmentlist_equal(c.known, a.known | b.known)

    def test_invert(self):
        """Test ``~a``."""
        a = self.TEST_CLASS(active=ACTIVE[:2], known=KNOWN)
        b = ~a
        utils.assert_segmentlist_equal(b.active, a.known & ~a.active)
        utils.assert_segmentlist_equal(b.known, a.known)

    def test_difference_simple(self):
        """Test that subtract works as intended for a simple case.

        Tests regression against https://github.com/gwpy/gwpy/issues/1700.
        """
        known1 = _as_segmentlist((0, 2), (3, 7))
        active1 = _as_segmentlist((1, 2), (3, 4), (5, 7))

        known2 = _as_segmentlist((3, 7), (8, 10))
        active2 = _as_segmentlist((4, 7), (9, 10))

        a = self.TEST_CLASS(active=active1, known=known1)
        b = self.TEST_CLASS(active=active2, known=known2)

        diff = a - b

        expected_known = _as_segmentlist((3, 7))
        expected_active = _as_segmentlist((3, 4))

        expected_diff = self.TEST_CLASS(
            active=expected_active,
            known=expected_known,
        )

        utils.assert_flag_equal(diff, expected_diff)

    def test_coalesce(self):
        """Test `DataQualityFlag.coalesce()`."""
        flag = self.create()
        flag.coalesce()
        utils.assert_segmentlist_equal(flag.known, KNOWN)
        utils.assert_segmentlist_equal(flag.active, KNOWNACTIVE)
        assert flag.regular is True

    def test_contract(self):
        """Test `DataQualityFlag.contract()`."""
        flag = self.create()
        flag.contract(.1)
        utils.assert_segmentlist_equal(flag.known, KNOWN)
        utils.assert_segmentlist_equal(flag.active, ACTIVE_CONTRACTED)

    def test_protract(self):
        """Test `DataQualityFlag.protract()`."""
        flag = self.create(active=ACTIVE_CONTRACTED)
        flag.protract(.1)
        utils.assert_segmentlist_equal(flag.known, KNOWN)
        utils.assert_segmentlist_equal(flag.active, ACTIVE)

    @pytest.mark.parametrize(("contract", "active"), [
        (False, ACTIVE_CONTRACTED),
        (True, ACTIVE_PROTRACTED),
    ])
    def test_round(self, contract, active):
        """Test `DataQualityFlag.round()`."""
        flag = self.create(active=active)
        r = flag.round(contract=contract)
        utils.assert_segmentlist_equal(r.known, KNOWN)
        utils.assert_segmentlist_equal(r.active, KNOWNACTIVE)

    def test_pad_noop(self, flag):
        """Test `DataQualityFlag.pad()`."""
        # test with no arguments (and no padding)
        padded = flag.pad()
        utils.assert_flag_equal(flag, padded)

    def test_pad_noargs(self, flag):
        """Test `DataQualityFlag.pad()` using ``flag.padding``."""
        flag.padding = PADDING
        padded = flag.pad()
        utils.assert_segmentlist_equal(padded.known, KNOWNPAD)
        utils.assert_segmentlist_equal(padded.active, ACTIVEPAD)

    def test_pad_args(self, flag):
        """Test `DataQualityFlag.pad()` with arguments."""
        flag.padding = (-100, 100)
        padded = flag.pad(*PADDING)
        utils.assert_segmentlist_equal(padded.known, KNOWNPAD)
        utils.assert_segmentlist_equal(padded.active, ACTIVEPAD)

    def test_pad_inplace(self, flag):
        """Test `DataQualityFlag.pad(..., inplace=...)`."""
        padded = flag.pad(*PADDING, inplace=False)
        assert padded is not flag

        padded = flag.pad(*PADDING, inplace=True)
        assert padded is flag
        utils.assert_segmentlist_equal(flag.known, KNOWNPAD)
        utils.assert_segmentlist_equal(flag.active, ACTIVEPAD)

        # check that other keyword arguments get rejected appropriately
        with pytest.raises(TypeError):
            flag.pad(*PADDING, kwarg="test")

    @pytest.mark.requires("igwn_ligolw.lsctables")
    def test_from_veto_def(self):
        """Test `DataQualityFlag.from_veto_def()`."""
        a = veto_def(
            "X1",
            "TEST-FLAG",
            1,
            start_time=0,
            end_time=0,
            start_pad=-2,
            end_pad=2,
            comment="Comment",
        )
        f = self.TEST_CLASS.from_veto_def(a)
        assert f.name == "X1:TEST-FLAG:1"
        assert f.category is None
        assert f.padding == (-2, 2)
        assert f.description == "Comment"
        utils.assert_segmentlist_equal(f.known, [(0, float("inf"))])

    @pytest.mark.requires("igwn_ligolw.lsctables")
    def test_from_veto_def_version(self):
        """Test `DataQualityFlag.from_veto_def()` with missing version."""
        a = veto_def("X1", "TEST-FLAG", None, start_time=0, end_time=1)
        f = self.TEST_CLASS.from_veto_def(a)
        assert f.name == "X1:TEST-FLAG"
        assert f.version is None

    @mock.patch("gwpy.segments.flag.query_segments", mock_query_segments)
    def test_populate(self):
        """Test `DataQualityFlag.populate()`."""
        name = QUERY_FLAGS[0]
        flag = self.TEST_CLASS(name, known=QUERY_RESULT[name].known)
        flag.populate()
        utils.assert_flag_equal(flag, QUERY_RESULTC[name])

    # -- test I/O --------------------

    @pytest.mark.parametrize(("format_", "ext", "rw_kwargs", "simple"), [
        ("hdf5", "hdf5", {"path": "test-dqflag"}, False),
        ("hdf5", "h5", {"path": "test-dqflag"}, False),
        ("json", "json", {}, True),
    ])
    def test_read_write(self, flag, format_, ext, rw_kwargs, simple):
        """Test writing and reading back a `DataQualityFlag`."""
        # simplify calling read/write tester
        def _read_write(**kwargs):
            read_kw = rw_kwargs.copy()
            read_kw.update(kwargs.pop("read_kw", {}))
            write_kw = rw_kwargs.copy()
            write_kw.update(kwargs.pop("write_kw", {}))
            return utils.test_read_write(
                flag,
                format_,
                extension=ext,
                assert_equal=utils.assert_flag_equal,
                read_kw=read_kw,
                write_kw=write_kw,
                **kwargs,
            )

        # perform simple test
        if simple:
            _read_write()

        # perform complicated test
        else:
            _read_write(autoidentify=False)
            with pytest.raises(
                OSError,
                match="test",
            ):
                _read_write(autoidentify=True)
            _read_write(autoidentify=True, write_kw={"overwrite": True})

    def test_read_write_hdf5(self, flag, tmp_path):
        """Test writing and reading back a `DataQualityFlag` in HDF5 format."""
        tmp = tmp_path / "test.h5"
        flag.write(tmp, path="test")
        f2 = self.TEST_CLASS.read(tmp, path="test", format="hdf5")
        utils.assert_flag_equal(f2, flag)

        # test direct access from dataset
        with h5py.File(tmp, "r") as h5f:
            f2 = self.TEST_CLASS.read(h5f["test"])
        utils.assert_flag_equal(f2, flag)

        # test auto-discover of single dataset
        f2 = self.TEST_CLASS.read(tmp)
        utils.assert_flag_equal(f2, flag)

    @pytest.mark.requires("igwn_ligolw.lsctables")
    def test_read_write_ligolw(self, flag):
        """Test writing and reading back a `DataQualityFlag` in LIGO_LW format."""
        utils.test_read_write(
            flag,
            "ligolw",
            extension="xml",
            assert_equal=utils.assert_flag_equal,
            autoidentify=False,
            read_kw={},
        )

    @pytest.mark.requires("igwn_ligolw.lsctables")
    def test_write_ligolw_attrs(self, tmp_path, flag):
        """Test writing a `DataQualityFlag` in LIGO_LW format with attributes."""
        from gwpy.io.ligolw import read_table
        tmp = tmp_path / "tmp.xml"
        flag.write(
            tmp,
            format="ligolw",
            attrs={"process_id": 100},
        )
        segdeftab = read_table(tmp, "segment_definer")
        assert int(segdeftab[0].process_id) == 100

    @pytest.mark.requires("igwn_ligolw.lsctables")
    def test_read_ligolw_no_ns(self, ligolw_no_ns):
        """Test writing and reading back a `DataQualityFlag` in LIGO_LW format.

        Specifically when the file doesn't populate nanosecond columns.
        """
        flag = self.TEST_CLASS.read(ligolw_no_ns, format="ligolw")
        assert flag.name == "X1:TEST_FLAG:1"
        assert flag.known == [(1366644592, 1366644608)]
        assert flag.active == [(1366644592, 1366644593)]

    # -- test queries ----------------

    @mock.patch("gwpy.segments.flag.query_segments", mock_query_segments)
    def test_query(self):
        """Test `DataQualityFlag.query()`."""
        result = self.TEST_CLASS.query(QUERY_FLAGS[0], 0, 10)
        assert isinstance(result, self.TEST_CLASS)
        expected = QUERY_RESULT[QUERY_FLAGS[0]].copy().coalesce()
        utils.assert_segmentlist_equal(result.known, expected.known)
        utils.assert_segmentlist_equal(result.active, expected.active)

    @pytest.mark.parametrize(("name", "flag"), [
        pytest.param(
            QUERY_FLAGS[0],
            QUERY_FLAGS[0],
            id="normal",
        ),
        pytest.param(
            QUERY_FLAGS[0].rsplit(":", 1)[0],
            QUERY_FLAGS[0],
            id="versionless",
        ),
    ])
    @mock.patch("gwpy.segments.flag.query_segments", mock_query_segments)
    def test_query_dqsegdb(self, name, flag):
        """Test `DataQualityFlag.query_dqsegdb()`."""
        # standard query
        result = self.TEST_CLASS.query_dqsegdb(name, 0, 10)
        expected = QUERY_RESULTC[flag]
        assert isinstance(result, self.TEST_CLASS)
        utils.assert_segmentlist_equal(result.known, expected.known)
        utils.assert_segmentlist_equal(result.active, expected.active)

        # segment as tuple
        result2 = self.TEST_CLASS.query_dqsegdb(name, (0, 10))
        utils.assert_flag_equal(result, result2)

        # segmentlist
        result2 = self.TEST_CLASS.query_dqsegdb(name, SegmentList([(0, 10)]))
        utils.assert_flag_equal(result, result2)

    def test_query_dqsegdb_error_flag_name(self):
        """Test `DataQualityFlag.query_dqsegdb()` flag name parsing error."""
        with pytest.raises(
            ValueError,
            match=r"Cannot parse ifo or tag \(name\) for flag 'BAD-FLAG_NAME'",
        ):
            self.TEST_CLASS.query_dqsegdb(
                "BAD-FLAG_NAME",
                SegmentList([(0, 10)]),
            )

    @mock.patch("gwpy.segments.flag.query_segments", mock_query_segments)
    def test_query_dqsegdb_not_found(self):
        """Test `DataQualityFlag.query_dqsegdb()` 404 annotation."""
        # flag not in database
        with pytest.raises(
            HTTPError,
            match=r"^404 Client Error: Not Found .* \[X1:GWPY-TEST:0\]$",
        ):
            self.TEST_CLASS.query_dqsegdb("X1:GWPY-TEST:0", 0, 10)

    @pytest.mark.parametrize("args", [
        (1, 2, 3),
        ((1, 2, 3),),
    ])
    def test_query_dqsegdb_args(self, args):
        """Test `DataQualityFlag.query_dqsegdb()` response to bad arguments."""
        # bad syntax
        with pytest.raises(
            ValueError,
            match=r"query_dqsegdb\(\) takes",
        ):
            self.TEST_CLASS.query_dqsegdb(QUERY_FLAGS[0], *args)

    @mock.patch("gwpy.segments.flag.query_segments", mock_query_segments)
    def test_query_dqsegdb_multi(self):
        """Test `DataQualityFlag.query_dqsegdb()` with multiple segments."""
        segs = SegmentList([Segment(0, 2), Segment(8, 10)])
        result = self.TEST_CLASS.query_dqsegdb(QUERY_FLAGS[0], segs)
        expected = QUERY_RESULTC[QUERY_FLAGS[0]]

        assert isinstance(result, self.TEST_CLASS)
        utils.assert_segmentlist_equal(result.known, expected.known & segs)
        utils.assert_segmentlist_equal(result.active, expected.active & segs)

    @pytest_skip_flaky_network
    def test_fetch_open_data(self):
        """Test `DataQualityFlag.fetch_open_data()`."""
        segs = self.TEST_CLASS.fetch_open_data(
            "H1_DATA",
            946339215,
            946368015,
        )
        assert segs.ifo == "H1"
        assert segs.name == "H1:DATA"
        assert segs.label == "H1_DATA"
        utils.assert_segmentlist_equal(segs.known, [(946339215, 946368015)])
        utils.assert_segmentlist_equal(segs.active, [
            (946340946, 946351799),
            (946356479, 946360619),
            (946362652, 946368015),
        ])


# -- DataQualityDict -----------------

class TestDataQualityDict:
    """Test `gwpy.segments.DataQualityDict."""

    TEST_CLASS = DataQualityDict
    ENTRY_CLASS = DataQualityFlag

    @classmethod
    def create(cls):
        """Create a `DataQualityDict`."""
        flgd = cls.TEST_CLASS()
        flgd["X1:TEST-FLAG:1"] = cls.ENTRY_CLASS(name="X1:TEST-FLAG:1",
                                                 active=ACTIVE, known=KNOWN)
        flgd["Y1:TEST-FLAG:2"] = cls.ENTRY_CLASS(name="Y1:TEST-FLAG:2",
                                                 active=ACTIVE2, known=KNOWN2)
        return flgd

    @pytest.fixture
    @classmethod
    def instance(cls):
        """Create a `DataQualityDict` (fixture)."""
        return cls.create()

    @pytest.fixture
    @classmethod
    def reverse(cls):
        """Create a `DataQualityDict` that is the reverse of ``instance``."""
        inst = cls.create()
        rev = type(inst)()
        keys = list(inst.keys())
        rev[keys[0]] = inst[keys[1]]
        rev[keys[1]] = inst[keys[0]]
        return rev

    # -- test logic ------------------

    def test_iand(self, instance, reverse):
        """Test ``a &= b``."""
        a = instance.copy()
        a &= reverse
        keys = list(a.keys())
        for key in keys:
            utils.assert_flag_equal(a[key], instance[key] & reverse[key])

    def test_and(self, instance, reverse):
        """Test ``a & b``."""
        a = instance.copy()
        a &= reverse
        utils.assert_dict_equal(a, instance & reverse, utils.assert_flag_equal)

    def test_ior(self, instance, reverse):
        """Test ``a |= b``."""
        a = instance.copy()
        a |= reverse
        keys = list(a.keys())
        for key in keys:
            utils.assert_flag_equal(a[key], instance[key] | reverse[key])

    def test_or(self, instance, reverse):
        """Test ``a | b``."""
        a = instance.copy()
        a |= reverse
        utils.assert_dict_equal(a, instance | reverse, utils.assert_flag_equal)

    def test_isub(self, instance, reverse):
        """Test ``a -= b``."""
        a = instance.copy()
        a -= reverse
        keys = list(a.keys())
        for key in keys:
            utils.assert_flag_equal(a[key], instance[key] - reverse[key])

    def test_sub(self, instance, reverse):
        """Test ``a - b``."""
        a = instance.copy(deep=True)
        a -= reverse
        utils.assert_dict_equal(a, instance - reverse, utils.assert_flag_equal)

    def test_ixor(self, instance, reverse):
        """Test ``a ^= b``."""
        a = instance.copy(deep=True)
        a ^= reverse
        utils.assert_dict_equal(a, instance ^ reverse, utils.assert_flag_equal)

    def test_invert(self, instance):
        """Test ``~a``."""
        inverse = type(instance)()
        for key in instance:
            inverse[key] = ~instance[key]
        utils.assert_dict_equal(~instance, inverse, utils.assert_flag_equal)

    # -- test methods ----------------

    def test_union(self, instance):
        """Test `DataQualityDict.union()`."""
        union = instance.union()
        assert isinstance(union, self.ENTRY_CLASS)
        utils.assert_segmentlist_equal(union.known, KNOWN + KNOWN2)
        utils.assert_segmentlist_equal(union.active, ACTIVE + ACTIVE2)

    def test_intersection(self, instance):
        """Test `DataQualityDict.intersection()`."""
        intersection = instance.intersection()
        assert isinstance(intersection, self.ENTRY_CLASS)
        utils.assert_segmentlist_equal(intersection.known, KNOWN & KNOWN2)
        utils.assert_segmentlist_equal(intersection.active, ACTIVE & ACTIVE2)

    def test_plot(self, instance):
        """Test `DataQualityDict.plot()`."""
        with rc_context(rc={"text.usetex": False}):
            plot = instance.plot(figsize=(6.4, 3.8))
            assert isinstance(plot.gca(), SegmentAxes)
            assert plot.get_figwidth() == 6.4
            assert plot.get_figheight() == 3.8
            plot.save(BytesIO(), format="png")
            plot.close()

    def test_plot_known_label(self, instance):
        """Test `DataQualityDict.plot(..., known=None, label=...)`."""
        with rc_context(rc={"text.usetex": False}):
            plot = instance.plot(
                known=None,
                figsize=(6.4, 3.8),
            )
            ax = plot.gca()
            for key, artist in zip(instance, ax.collections, strict=True):
                assert artist.get_label() == instance[key].name
            plot.save(BytesIO(), format="png")
            plot.close()

    def test_plot_label(self, instance):
        """Test `DataQualityDict.plot(..., label=...)`."""
        with rc_context(rc={"text.usetex": False}):
            plot = instance.plot(
                label="Fixed label",
                figsize=(6.4, 3.8),
            )
            ax = plot.gca()
            for artist in ax.collections:
                assert artist.get_label() == "Fixed label"
            plot.save(BytesIO(), format="png")
            plot.close()

    # -- test I/O --------------------

    @pytest.mark.requires("igwn_ligolw.lsctables")
    def test_from_veto_definer_file(self, veto_definer):
        """Test `DataQualityDict.from_veto_definer_file()`."""
        # read veto definer
        vdf = self.TEST_CLASS.from_veto_definer_file(veto_definer)
        assert len(vdf.keys()) == 4

        # test one flag to make sure it is well read
        name = "X1:TEST-FLAG:1"
        assert name in vdf
        utils.assert_segmentlist_equal(vdf[name].known,
                                       [(100, float("inf"))])
        assert vdf[name].category == 1
        assert vdf[name].padding == (-1, 2)

        # test ifo kwarg
        vdf = self.TEST_CLASS.from_veto_definer_file(veto_definer, ifo="X1")
        assert len(vdf.keys()) == 3
        assert "Y1:TEST-FLAG_2:2" not in vdf

        # test start and end kwargs
        vdf = self.TEST_CLASS.from_veto_definer_file(veto_definer,
                                                     start=200, end=300)
        assert len(vdf.keys()) == 3
        assert "X1:TEST-FLAG_2:1" not in vdf

    @pytest.mark.parametrize(("format_", "ext", "rw_kwargs"), [
        ("hdf5", "hdf5", {}),
        ("hdf5", "h5", {}),
        ("hdf5", "hdf5", {"path": "test-dqdict"}),
    ])
    def test_read_write(self, instance, format_, ext, rw_kwargs):
        """Test writing and reading back a `DataQualityDict`."""
        # define assertion
        def _assert(a, b):
            return utils.assert_dict_equal(a, b, utils.assert_flag_equal)

        # simplify calling read/write tester
        def _read_write(**kwargs):
            read_kw = rw_kwargs.copy()
            read_kw.update(kwargs.pop("read_kw", {}))
            write_kw = rw_kwargs.copy()
            write_kw.update(kwargs.pop("write_kw", {}))
            return utils.test_read_write(
                instance,
                format_,
                extension=ext,
                assert_equal=_assert,
                read_kw=read_kw,
                write_kw=write_kw,
                **kwargs,
            )

        _read_write(autoidentify=False)

        # test that attempting to read/write again fails
        with pytest.raises(
            OSError,
            match="File exists:",
        ):
            _read_write(autoidentify=True)

        # unless overwrite=True is given
        _read_write(autoidentify=True, write_kw={"overwrite": True})

    @pytest.mark.requires("igwn_ligolw.lsctables")
    def test_read_write_ligolw(self, instance):
        """Test writing and reading back a `DataQualityDict` in LIGO_LW format."""
        def _assert(a, b):
            return utils.assert_dict_equal(a, b, utils.assert_flag_equal)

        utils.test_read_write(
            instance,
            "ligolw",
            extension="xml",
            assert_equal=_assert,
            autoidentify=False,
            read_kw={},
        )

    def test_read_on_missing(self, instance):
        """Test `DataQualityDict.read(..., on_missing=...)`."""
        with h5py.File(
                "test",
                mode="w-",
                driver="core",
                backing_store=False,
        ) as h5f:
            instance.write(h5f)
            names = ["randomname"]

            def _read(**kwargs):
                return self.TEST_CLASS.read(
                    h5f,
                    names=names,
                    format="hdf5",
                    **kwargs,
                )

            # check on_missing='error' (default) raises ValueError
            with pytest.raises(
                ValueError,
                match=r"^'randomname' not found in any input file$",
            ):
                _read()

            # check on_missing='warn' prints warning
            with pytest.warns(UserWarning):
                _read(on_missing="warn")

            # check on_missing='ignore' does nothing
            with warnings.catch_warnings():
                warnings.simplefilter("error")
                _read(on_missing="ignore")

            # check on_missing=<anything else> raises exception
            with pytest.raises(
                ValueError,
                match=r"^'randomname' not found in any input file$",
            ):
                _read(on_missing="blah")

    @pytest.mark.requires("igwn_ligolw.lsctables")
    def test_to_ligolw_tables(self, instance):
        """Test `DataQualityDict.to_ligolw_tables()`."""
        tables = instance.to_ligolw_tables()
        assert len(tables[0]) == len(instance)  # segdef
        assert len(tables[1]) == sum(len(x.known) for x in instance.values())
        assert len(tables[2]) == sum(len(x.active) for x in instance.values())

    # -- test queries ----------------

    @mock.patch("gwpy.segments.flag.query_segments", mock_query_segments)
    def test_query(self):
        """Test `DataQualityDict.query()`."""
        result = self.TEST_CLASS.query(QUERY_FLAGS, 0, 10)
        expected = QUERY_RESULT.copy().coalesce()

        assert isinstance(result, self.TEST_CLASS)
        utils.assert_dict_equal(result, expected, utils.assert_flag_equal)

    @mock.patch("gwpy.segments.flag.query_segments", mock_query_segments)
    def test_query_dqsegdb(self):
        """Test `DataQualityDict.query_dqsegdb()`."""
        result = self.TEST_CLASS.query_dqsegdb(QUERY_FLAGS, 0, 10)
        assert isinstance(result, self.TEST_CLASS)
        utils.assert_dict_equal(
            result,
            QUERY_RESULTC,
            utils.assert_flag_equal,
        )

    @pytest.mark.parametrize(("on_error", "ctx"), [
        # does nothing
        ("ignore", nullcontext()),
        # emits a warning
        ("warn", pytest.warns(UserWarning)),
        # propagates the exception from the thread
        ("raise", pytest.raises(
            HTTPError,
            match=r"404 Client Error: Not Found for .* \[X1:BLAHBLAH:1\]",
        )),
        # invalid value
        ("blah", pytest.raises(ValueError, match="on_error must be one of")),
    ])
    @mock.patch("gwpy.segments.flag.query_segments", mock_query_segments)
    def test_query_dqsegdb_on_error(self, on_error, ctx):
        """Test `DataQualityDict.query_dqsegdb(..., on_error=...)`."""
        with ctx as record:
            result = self.TEST_CLASS.query_dqsegdb(
                [*QUERY_FLAGS, "X1:BLAHBLAH:1"],
                0,
                10,
                on_error=on_error,
                nthreads=2,
            )
        if on_error == "warn":
            assert len(record) == 1  # check on_error='ignore' didn't warn
        elif on_error != "ignore":
            return  # exception has been asserted
        assert isinstance(result, self.TEST_CLASS)
        utils.assert_dict_equal(
            result,
            QUERY_RESULTC,
            utils.assert_flag_equal,
        )

    @mock.patch("gwpy.segments.flag.query_segments", mock_query_segments)
    def test_populate(self):
        """Test `DataQualityDict.populate()`."""
        def fake():
            return self.TEST_CLASS({
                x: self.ENTRY_CLASS(name=x, known=y.known)
                for x, y in QUERY_RESULT.items()
            })

        # build fake veto definer file
        vdf = fake()
        vdf2 = fake()
        vdf3 = fake()

        flag = QUERY_FLAGS[0]
        vdf2[flag].padding = (-1, 1)

        span = SegmentList([Segment(0, 2)])

        # and populate using a mocked query
        vdf.populate()
        vdf2.populate()
        vdf3.populate(segments=span)

        # test warnings on bad entries
        vdf["TEST"] = self.ENTRY_CLASS("X1:BLAHBLAHBLAH:1", known=[(0, 1)])
        with pytest.warns(UserWarning) as record:
            vdf.populate(on_error="warn")
            vdf.populate(on_error="ignore")
        assert len(record) == 1
        vdf.pop("TEST")

        with pytest.raises(
            ValueError,
            match="on_error must be one of",
        ):
            vdf.populate(on_error="blah")

        # check basic populate worked
        utils.assert_dict_equal(vdf, QUERY_RESULTC, utils.assert_flag_equal)

        # check padded populate worked
        utils.assert_flag_equal(vdf2[flag], QUERY_RESULTC[flag].pad(-1, 1))

        # check segment-restricted populate worked
        for flag in vdf3:
            utils.assert_segmentlist_equal(
                vdf3[flag].known, QUERY_RESULTC[flag].known & span)
            utils.assert_segmentlist_equal(
                vdf3[flag].active, QUERY_RESULTC[flag].active & span)

    def test_coalesce(self):
        """Test `DataQualityDict.coalesce()`."""
        instance = self.create()
        instance.coalesce()
        value = instance["X1:TEST-FLAG:1"]
        utils.assert_segmentlist_equal(value.active, KNOWNACTIVE)



# Copyright (c) 2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Colour customisations for visualisation in GWpy."""

from __future__ import annotations

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

from typing import TYPE_CHECKING

import numpy
from matplotlib.colors import (
    LogNorm,
    Normalize,
    get_named_colors_mapping,
    hsv_to_rgb,
    rgb_to_hsv,
    to_rgb,
)

if TYPE_CHECKING:
    from typing import Any

    from matplotlib.typing import (
        ColorType,
        RGBColorType,
    )

# -- recommended defaults for current Gravitational-Wave Observatories
# the below colours are designed to work well for the colour-blind, as well
# as in grayscale, so are recommended for publications

GWPY_COLORS = {
    "geo600": "#222222",  # dark gray
    "kagra": "#ffb200",  # yellow/orange
    "ligo-hanford": "#ee0000",  # red
    "ligo-india": "#b0dd8b",  # light green
    "ligo-livingston": "#4ba6ff",  # blue
    "virgo": "#9b59b6",  # magenta/purple
}

# provide user mapping by IFO prefix
_GWO_PREFIX = {
    "geo600": "G1",
    "kagra": "K1",
    "ligo-hanford": "H1",
    "ligo-india": "I1",
    "ligo-livingston": "L1",
    "virgo": "V1",
}
GW_OBSERVATORY_COLORS = {_GWO_PREFIX[n]: GWPY_COLORS[n] for n in GWPY_COLORS}

# set named colour upstream in matplotlib, so users can specify as
# e.g. plot(..., color='gwpy:ligo-hanford')           noqa: ERA001
get_named_colors_mapping().update({
    f"gwpy:{name}": col for name, col in GWPY_COLORS.items()
})


# -- colour utilities ---------------------------------------------------------

def tint(col: ColorType, factor: float = 1.0) -> RGBColorType:
    """Tint a color (make it darker), returning a new RGB array.

    Parameters
    ----------
    col : `matplotlib.typing.ColorType`
        The colour to darken.

    factor : `float`, optional
        The amount by which to darken it. Values less than ``1`` will make things
        darker (``0`` will return black), while less than ``1`` will make things
        lighter.

    Returns
    -------
    tinted : `matplotlib.typing.RGBColorType`
        The new RGB colour tuple.
    """
    # this method is more complicated than it need be to
    # support matplotlib-1.x.
    # for matplotlib-2.x this would just be
    #     h, s, v = rgb_to_hsv(to_rgb(c))  noqa: ERA001
    #     v *= factor                      noqa: ERA001
    #     return hsv_to_rgb((h, s, v))     noqa: ERA001
    rgb = numpy.array(to_rgb(col), ndmin=3)
    hsv = rgb_to_hsv(rgb)
    hsv[-1][-1][2] *= factor
    return hsv_to_rgb(hsv)[-1][-1]


def format_norm(
    kwargs: dict[str, Any],
    current: str | Normalize | None = None,
) -> tuple[Normalize, dict[str, Any]]:
    """Format a `~matplotlib.colors.Normalize` from a set of kwargs.

    Parameters
    ----------
    kwargs : `dict`
        A set of keyword arguments to a plotting routine.

    current : `Normalize`, optional
        The current normalization, which will be preserved if
        ``norm`` is not specified in ``kwargs``.

    Returns
    -------
    norm, kwargs
        The formatted `Normalize` instance, and the remaining keywords.

    Raises
    ------
    ValueError
        If ``"norm"`` is included in ``kwargs`` and doesn't match a
        recognised value (one of ``"linear"``, ``"log"`` or a
        `~matplotlib.colors.Normalize` instance).
    """
    norm = kwargs.pop("norm", current) or "linear"
    vmin = kwargs.pop("vmin", None)
    vmax = kwargs.pop("vmax", None)
    clim = kwargs.pop("clim", (vmin, vmax)) or (None, None)
    clip = kwargs.pop("clip", None)

    if norm == "linear":
        norm = Normalize()
    elif norm == "log":
        norm = LogNorm()
    elif not isinstance(norm, Normalize):
        msg = f"unrecognised value for norm '{norm}'"
        raise ValueError(msg)

    for attr, value in (
        ("vmin", clim[0]),
        ("vmax", clim[1]),
        ("clip", clip),
    ):
        if value is not None:
            setattr(norm, attr, value)

    return norm, kwargs



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWpy API to the LAL FFT routines.

See the `LAL TimeFreqFFT.h documentation
<https://lscsoft.docs.ligo.org/lalsuite/lal/group___time_freq_f_f_t__h.html>`_
for more details

This module is deprecated and will be removed in a future release.
"""

from __future__ import annotations

import re
import warnings
from typing import (
    TYPE_CHECKING,
    Literal,
    cast,
)

import numpy
from astropy import units

from ...frequencyseries import FrequencySeries
from ..window import canonical_name
from . import _registry as fft_registry
from ._utils import scale_timeseries_unit

if TYPE_CHECKING:
    from numpy.typing import DTypeLike

    from ...timeseries import TimeSeries
    from ...utils.lal import (
        LALFFTPlanType,
        LALWindowType,
    )
    from ..window import WindowLike

    LALAverageSpectrumMethods = Literal[
        "AverageSpectrumMedianMean",
        "AverageSpectrumMedian",
        "AverageSpectrumWelch",
    ]

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# cache windows and FFT plans internally
LAL_WINDOWS: dict[tuple[int, str, str], LALWindowType] = {}
LAL_FFTPLANS: dict[tuple[int, bool, str], LALFFTPlanType] = {}

#: Default FFT plan level for LAL FFT plans
LAL_FFTPLAN_LEVEL = 1


# -- utilities -----------------------

def generate_fft_plan(
    length: int,
    level: int | None = None,
    dtype: DTypeLike = "float64",
    *,
    forward: bool = True,
) -> LALFFTPlanType:
    """Build a `REAL8FFTPlan` for a fast Fourier transform.

    Parameters
    ----------
    length : `int`
        Number of samples to plan for in each FFT.

    level : `int`, optional
        Amount of work to do when planning the FFT, default set by
        `LAL_FFTPLAN_LEVEL` module variable.

    dtype : :class:`numpy.dtype`, `type`, `str`, optional
        Numeric type of data to plan for

    forward : bool, optional, default: `True`
        Whether to create a forward or reverse FFT plan.

    Returns
    -------
    plan : `REAL8FFTPlan` or similar
        FFT plan of the relevant data type.
    """
    from ...utils.lal import find_typed_function, to_lal_type_str

    # generate key for caching plan
    laltype = to_lal_type_str(dtype)
    key = (length, bool(forward), laltype)

    # find existing plan
    try:
        return LAL_FFTPLANS[key]
    # or create one
    except KeyError:
        create = find_typed_function(dtype, "Create", "FFTPlan")
        if level is None:
            level = LAL_FFTPLAN_LEVEL
        LAL_FFTPLANS[key] = create(length, int(bool(forward)), level)
        return LAL_FFTPLANS[key]


def generate_window(
    length: int,
    window: WindowLike | None = None,
    dtype: DTypeLike | None = "float64",
) -> LALWindowType:
    """Generate a time-domain window for use in a LAL FFT.

    Parameters
    ----------
    length : `int`
        Length of window in samples.

    window : `str`, `tuple`
        Name of window to generate, default: ``('kaiser', 24)``. Give
        `str` for simple windows, or tuple of ``(name, *args)`` for
        complicated windows.

    dtype : `numpy.dtype`
        Numeric type of window, default ``numpy.dtype(numpy.float64)``.

    Returns
    -------
    `window` : `REAL8Window` or similar
        Time-domain window to use for FFT.
    """
    from ...utils.lal import find_typed_function, to_lal_type_str

    if window is None:
        window = ("kaiser", 24)

    if dtype is None:
        dtype = float

    # generate key for caching window
    laltype = to_lal_type_str(dtype)
    key = (length, str(window), laltype)

    # find existing window
    try:
        return LAL_WINDOWS[key]
    # or create one
    except KeyError:
        # handle arrays directly
        if isinstance(window, numpy.ndarray):
            return window_from_array(window, dtype=dtype)
        # parse window as name and arguments, e.g. ('kaiser', 24)
        if isinstance(window, list | tuple):
            window, beta = window
        else:
            beta = 0
        window = canonical_name(str(window))
        # create window
        create = find_typed_function(dtype, "CreateNamed", "Window")
        LAL_WINDOWS[key] = create(window, beta, length)
        return LAL_WINDOWS[key]


def window_from_array(
    array: numpy.ndarray,
    dtype: DTypeLike | None = None,
) -> LALWindowType:
    """Convert a `numpy.ndarray` into a LAL `Window` object."""
    from ...utils.lal import find_typed_function

    if dtype is None:
        dtype = array.dtype

    # create sequence
    seq = find_typed_function(dtype, "Create", "Sequence")(array.size)
    seq.data = numpy.asarray(array, dtype=dtype)

    # create window from sequence
    return find_typed_function(dtype, "Create", "WindowFromSequence")(seq)


# -- spectrumm methods ---------------

def _lal_spectrum(
    timeseries: TimeSeries,
    segmentlength: int,
    noverlap: int | None = None,
    method: str = "welch",
    window: LALWindowType | WindowLike | None = None,
    plan: LALFFTPlanType | None = None,
) -> FrequencySeries:
    """Generate a PSD `FrequencySeries` using |lal|_.

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        Input `TimeSeries` data.

    segmentlength : `int`
        Number of samples in single average.

    method : `str`
        Average PSD method.

    noverlap : `int`
        Number of samples to overlap between segments, defaults to 50%.

    window : `lal.REAL8Window`, optional
        Window to apply to timeseries prior to FFT.

    plan : `lal.REAL8FFTPlan`, optional
        LAL FFT plan to use when generating average spectrum.

    Returns
    -------
    spectrum : `~gwpy.frequencyseries.FrequencySeries`
        Average power `FrequencySeries`.
    """
    import lal

    from ...utils.lal import (
        LALWindowType,
        find_typed_function,
    )

    # default to 50% overlap
    if noverlap is None:
        noverlap = int(segmentlength // 2)
    stride = segmentlength - noverlap

    # get window
    if not isinstance(window, LALWindowType):
        window = generate_window(
            segmentlength,
            window=window,
            dtype=timeseries.dtype,
        )

    # get FFT plan
    if plan is None:
        plan = generate_fft_plan(segmentlength, dtype=timeseries.dtype)

    method = method.lower()

    # check data length
    size = timeseries.size
    numsegs = 1 + int((size - segmentlength) / stride)
    if method == "median-mean" and numsegs % 2:
        numsegs -= 1
        if not numsegs:
            msg = "Cannot calculate median-mean spectrum with this small a TimeSeries."
            raise ValueError(msg)

    required = int((numsegs - 1) * stride + segmentlength)
    if size != required:
        warnings.warn(
            "Data array is the wrong size for the correct number "
            "of averages given the input parameters. The trailing "
            f"{size - required} samples will not be used in this calculation.",
            stacklevel=2,
        )
        timeseries = timeseries[:required]

    # generate output spectrum
    create = find_typed_function(timeseries.dtype, "Create", "FrequencySeries")
    lalfs = create(
        timeseries.name or "",
        lal.LIGOTimeGPS(timeseries.t0.value),
        0,
        1 / segmentlength,
        lal.StrainUnit,
        int(segmentlength // 2 + 1),
    )

    # find LAL method (e.g. median-mean -> lal.REAL8AverageSpectrumMedianMean)
    methodname = cast(
        "LALAverageSpectrumMethods",
        f"AverageSpectrum{''.join(map(str.title, re.split('[-_]', method)))}",
    )
    spec_func = find_typed_function(
        timeseries.dtype,
        "",
        methodname,
    )

    # calculate spectrum
    spec_func(lalfs, timeseries.to_lal(), segmentlength, stride, window, plan)

    # format and return
    spec = FrequencySeries.from_lal(lalfs)
    spec.name = timeseries.name
    spec.channel = timeseries.channel
    spec.override_unit(
        scale_timeseries_unit(
            timeseries.unit or units.dimensionless_unscaled,
            scaling="density",
        ),
    )
    return spec


def welch(
    timeseries: TimeSeries,
    segmentlength: int,
    noverlap: int | None = None,
    window: WindowLike | None = None,
    plan: LALFFTPlanType | None = None,
) -> FrequencySeries:
    """Calculate an PSD of this `TimeSeries` using Welch's method.

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        Input `TimeSeries` data.

    segmentlength : `int`
        Number of samples in single average.

    noverlap : `int`
        Number of samples to overlap between segments, defaults to 50%.

    window : `tuple`, `str`, optional
        Window parameters to apply to timeseries prior to FFT.

    plan : `REAL8FFTPlan`, optional
        LAL FFT plan to use when generating average spectrum.

    Returns
    -------
    spectrum : `~gwpy.frequencyseries.FrequencySeries`
        Average power `FrequencySeries`.

    See Also
    --------
    lal.REAL8AverageSpectrumWelch
    """
    return _lal_spectrum(
        timeseries,
        segmentlength,
        noverlap=noverlap,
        method="welch",
        window=window,
        plan=plan,
    )


def bartlett(
    timeseries: TimeSeries,
    segmentlength: int,
    noverlap: int | None = None,  # noqa: ARG001
    window: WindowLike | None = None,
    plan: LALFFTPlanType | None = None,
) -> FrequencySeries:
    # pylint: disable=unused-argument
    """Calculate an PSD of this `TimeSeries` using Bartlett's method.

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        Input `TimeSeries` data.

    segmentlength : `int`
        Number of samples in single average.

    noverlap : `int`
        Number of samples to overlap between segments, defaults to 50%.

    window : `tuple`, `str`, optional
        Window parameters to apply to timeseries prior to FFT.

    plan : `REAL8FFTPlan`, optional
        LAL FFT plan to use when generating average spectrum.

    Returns
    -------
    spectrum : `~gwpy.frequencyseries.FrequencySeries`
        Average power `FrequencySeries`

    See Also
    --------
    lal.REAL8AverageSpectrumWelch
    """
    return _lal_spectrum(
        timeseries,
        segmentlength,
        noverlap=0,
        method="welch",
        window=window,
        plan=plan,
    )


def median(
    timeseries: TimeSeries,
    segmentlength: int,
    noverlap: int | None = None,
    window: WindowLike | None = None,
    plan: LALFFTPlanType | None = None,
) -> FrequencySeries:
    """Calculate a PSD of this `TimeSeries` using a median average method.

    The median average is similar to Welch's method, using a median average
    rather than mean.

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        Input `TimeSeries` data.

    segmentlength : `int`
        Number of samples in single average.

    noverlap : `int`
        Number of samples to overlap between segments, defaults to 50%.

    window : `tuple`, `str`, optional
        Window parameters to apply to timeseries prior to FFT.

    plan : `REAL8FFTPlan`, optional
        LAL FFT plan to use when generating average spectrum.

    Returns
    -------
    spectrum : `~gwpy.frequencyseries.FrequencySeries`
        Average power `FrequencySeries`.

    See Also
    --------
    lal.REAL8AverageSpectrumMedian
    """
    return _lal_spectrum(
        timeseries,
        segmentlength,
        noverlap=noverlap,
        method="median",
        window=window,
        plan=plan,
    )


def median_mean(
    timeseries: TimeSeries,
    segmentlength: int,
    noverlap: int | None = None,
    window: WindowLike | None = None,
    plan: LALFFTPlanType | None = None,
) -> FrequencySeries:
    """Calculate a PSD of this `TimeSeries` using a median-mean average method.

    The median-mean average method divides overlapping segments into "even"
    and "odd" segments, and computes the bin-by-bin median of the "even"
    segments and the "odd" segments, and then takes the bin-by-bin average
    of these two median averages.

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        Input `TimeSeries` data.

    segmentlength : `int`
        Number of samples in single average.

    noverlap : `int`
        Number of samples to overlap between segments, defaults to 50%.

    window : `tuple`, `str`, optional
        Window parameters to apply to timeseries prior to FFT.

    plan : `REAL8FFTPlan`, optional
        LAL FFT plan to use when generating average spectrum.

    Returns
    -------
    spectrum : `~gwpy.frequencyseries.FrequencySeries`
        Average power `FrequencySeries`.

    See Also
    --------
    lal.REAL8AverageSpectrumMedianMean
    """
    return _lal_spectrum(
        timeseries,
        segmentlength,
        noverlap=noverlap,
        method="median-mean",
        window=window,
        plan=plan,
    )


# register LAL methods without overriding scipy method
for func in (welch, bartlett, median, median_mean):
    fft_registry.register_method(
        func,
        name=f"lal-{func.__name__}",
        deprecated=True,
    )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for frequencyseries module."""

from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

import numpy
import pytest
from astropy import units
from numpy import shares_memory
from numpy.testing import assert_allclose
from scipy import signal

from ...signal import filter_design
from ...testing import utils
from ...timeseries import TimeSeries
from ...types.tests.test_series import TestSeries as _TestSeries
from .. import FrequencySeries

if TYPE_CHECKING:
    from pathlib import Path

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

LIGO_LW_ARRAY = r"""<?xml version='1.0' encoding='utf-8'?>
<!DOCTYPE LIGO_LW SYSTEM "http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt">
<LIGO_LW>
  <LIGO_LW Name="REAL8FrequencySeries">
    <Time Type="GPS" Name="epoch">1000000000</Time>
    <Param Type="lstring" Name="channel:param">X1:TEST-CHANNEL_1</Param>
    <Array Type="real_8" Name="PSD1:array" Unit="Hz^-1">
      <Dim Start="10" Scale="1" Name="Frequency" Unit="Hz">5</Dim>
      <Dim Name="Frequency,Real">2</Dim>
      <Stream Delimiter=" " Type="Local">
        0 1
        1 2
        2 3
        3 4
        4 5
      </Stream>
    </Array>
  </LIGO_LW>
  <LIGO_LW Name="REAL8FrequencySeries">
    <Param Type="lstring" Name="channel:param">X1:TEST-CHANNEL_2</Param>
    <Param Type="real_8" Name="f0:param" Unit="s^-1">0</Param>
    <Array Type="real_8" Name="PSD2:array" Unit="s m^2">
      <Dim Start="0" Scale="1" Name="Frequency" Unit="s^-1">5</Dim>
      <Dim Name="Real">1</Dim>
      <Stream Delimiter=" " Type="Local">
        10
        20
        30
        40
        50
      </Stream>
    </Array>
  </LIGO_LW>
  <LIGO_LW Name="REAL8FrequencySeries">
    <Time Type="GPS" Name="epoch">1000000001</Time>
    <Array Type="real_8" Name="PSD2:array" Unit="s m^2">
      <Dim Start="0" Scale="1" Name="Frequency" Unit="s^-1">5</Dim>
      <Dim Name="Frequency,Real">3</Dim>
      <Stream Delimiter=" " Type="Local">
        0 10 1
        1 20 2
        2 30 3
        3 40 4
        4 50 5
      </Stream>
    </Array>
  </LIGO_LW>
</LIGO_LW>
"""


class TestFrequencySeries(_TestSeries):
    """Tests for `FrequencySeries`."""

    TEST_CLASS = FrequencySeries

    # -- test properties -------------

    def test_f0(self, array):
        """Test `FrequencySeries.f0`."""
        assert array.f0 is array.x0
        array.f0 = 4
        assert array.f0 == 4 * units.Hz

    def test_df(self, array):
        """Test `FrequencySeries.df`."""
        assert array.df is array.dx
        array.df = 4
        assert array.df == 4 * units.Hz

    def test_frequencies(self, array):
        """Test `FrequencySeries.frequencies`."""
        assert array.frequencies is array.xindex
        utils.assert_quantity_equal(
            array.frequencies, numpy.arange(array.size) * array.df + array.f0)

    # -- test methods ----------------

    @pytest.mark.usefixtures("usetex")
    def test_plot(self, array):
        """Test `FrequencySeries.plot`."""
        plot = array.plot()
        line = plot.gca().lines[0]
        utils.assert_array_equal(line.get_xdata(), array.xindex.value)
        utils.assert_array_equal(line.get_ydata(), array.value)
        plot.save(BytesIO(), format="png")
        plot.close()

    def test_ifft(self):
        """Test `FrequencySeries.ifft`."""
        # construct a TimeSeries, then check that it is unchanged by
        # the operation TimeSeries.fft().ifft()
        ts = TimeSeries([1.0, 0.0, -1.0, 0.0], sample_rate=1.0)
        utils.assert_quantity_sub_equal(ts.fft().ifft(), ts)
        utils.assert_allclose(ts.fft().ifft().value, ts.value)

    @pytest.mark.parametrize(("filter_form", "analog"), [
        pytest.param("zpk", True, id="zpk-analog"),
        pytest.param("zpk", False, id="zpk-digital"),
        pytest.param("ba", True, id="ba-analog"),
        pytest.param("ba", False, id="ba-digital"),
        pytest.param("sos", False, id="sos-digital"),
    ])
    def test_filter(
        self,
        array: FrequencySeries,
        filter_form: str,
        analog: bool,
    ):
        """Test `FrequencySeries.filter()`."""
        # Design a lowpass filter at nyquist/2
        nyq = array.frequencies.value[-1]
        filt_kw = {
            "btype": "low",
            "analog": analog,
            "output": filter_form,
        }
        if not analog:
            filt_kw["fs"] = nyq * 2
        filt = signal.butter(3, nyq / 2, **filt_kw)

        # Apply the filter
        a2 = array.filter(filt, analog=analog)

        # Check that we get back a FrequencySeries with the correct frequencies
        assert isinstance(a2, type(array))
        utils.assert_quantity_equal(a2.frequencies, array.frequencies)

        # Manually compute the filter response
        _, fresp = numpy.abs(filter_design.frequency_response(
            filt,
            array.frequencies.to("Hz").value,
            analog=analog,
            sample_rate=nyq * 2,
            unit="rad/s",
        ))

        # Check that the filter was applied correctly
        assert_allclose(a2.value, fresp * array.value)

    def test_zpk_analog_hertz(self, array):
        """Test `FrequencySeries.zpk(..., analog=True)`."""
        from gwpy.signal.filter_design import _convert_zpk_units

        zpk = ([100], [1], 1e-2)
        a2 = array.zpk(*zpk, unit="Hertz", analog=True)

        # Rebuild frequency-response manually
        z_rad, p_rad, k_rad = _convert_zpk_units(zpk, "Hz")
        omega = array.frequencies.value * 2 * numpy.pi
        fresp = numpy.abs(signal.freqs_zpk(z_rad, p_rad, k_rad, omega)[1])

        utils.assert_array_equal(a2.value, fresp * array.value)

    def test_inject(self):
        """Test `FrequencySeries.inject`."""
        # create a timeseries out of an array of zeros
        df, nyquist = 1, 2048
        nsamp = int(nyquist/df) + 1
        data = FrequencySeries(numpy.zeros(nsamp), f0=0, df=df, unit="")

        # create a second timeseries to inject into the first
        w_nyquist = 1024
        w_nsamp = int(w_nyquist/df) + 1
        sig = FrequencySeries(numpy.ones(w_nsamp), f0=0, df=df, unit="")

        # test that we recover this waveform when we add it to data,
        # and that the operation does not change the original data
        new_data = data.inject(sig)
        assert new_data.unit == data.unit
        assert new_data.size == data.size
        ind, = new_data.value.nonzero()
        assert len(ind) == sig.size
        utils.assert_allclose(new_data.value[ind], sig.value)
        utils.assert_allclose(data.value, numpy.zeros(nsamp))

    def test_interpolate(self):
        """Test `FrequencySeries.interpolate`."""
        # create a simple FrequencySeries
        df, nyquist = 1, 256
        nsamp = int(nyquist/df) + 1
        fseries = FrequencySeries(numpy.ones(nsamp), f0=1, df=df, unit="")

        # create an interpolated FrequencySeries
        newf = fseries.interpolate(df/2.)

        # check that the interpolated series is what was expected
        assert newf.unit == fseries.unit
        assert newf.size == 2*(fseries.size - 1) + 1
        assert newf.df == fseries.df / 2.
        assert newf.f0 == fseries.f0
        utils.assert_allclose(newf.value, numpy.ones(2*int(nyquist/df) + 1))

    @pytest.mark.requires("lal")
    def test_to_from_lal(self, array):
        """Test `FrequencySeries.to_lal` and `FrequencySeries.from_lal`."""
        import lal

        array.epoch = 0

        # check that to + from returns the same array
        lalts = array.to_lal()
        a2 = type(array).from_lal(lalts)
        utils.assert_quantity_sub_equal(array, a2, exclude=["channel"])

        # test copy=False
        a2 = type(array).from_lal(lalts, copy=False)
        assert shares_memory(a2.value, lalts.data.data)

        # test units
        array.override_unit("undef")
        with pytest.warns(
            UserWarning,
            match="LAL has no unit corresponding to 'NONE'",
        ):
            lalts = array.to_lal()
        assert lalts.sampleUnits == lal.DimensionlessUnit
        a2 = self.TEST_CLASS.from_lal(lalts)
        assert a2.unit == units.dimensionless_unscaled

    def test_to_from_pycbc(self, array):
        """Test `FrequencySeries.to_pycbc` and `FrequencySeries.from_pycbc`."""
        pycbctypes = pytest.importorskip("pycbc.types")

        array.epoch = 0

        # test default conversion
        pycbcfs = array.to_pycbc()
        assert isinstance(pycbcfs, pycbctypes.FrequencySeries)
        utils.assert_array_equal(array.value, pycbcfs.data)
        assert array.f0.value == 0 * units.Hz
        assert array.df.value == pycbcfs.delta_f
        assert array.epoch.gps == pycbcfs.epoch

        # go back and check we get back what we put in in the first place
        a2 = type(array).from_pycbc(pycbcfs)
        utils.assert_quantity_sub_equal(
            array, a2, exclude=["name", "unit", "channel"])

        # test copy=False
        a2 = type(array).from_pycbc(array.to_pycbc(copy=False), copy=False)
        assert shares_memory(array.value, a2.value)

    @pytest.mark.requires("pycbc")
    def test_to_from_pycbc_nonzero_f0(self, array):
        """Test `FrequencySeries.to_pycbc` conversion when ``f0 != 0``."""
        array.f0 = 1.
        with pytest.raises(
            ValueError,
            match=r"^Cannot convert FrequencySeries",
        ):
            array.to_pycbc()

    @pytest.mark.parametrize("fmt", [
        "txt",
        "csv",
    ])
    def test_read_write(self, array, fmt):
        """Test reading and writing FrequencySeries in various formats."""
        utils.test_read_write(
            array,
            fmt,
            assert_equal=utils.assert_quantity_sub_equal,
            assert_kw={
                "exclude": ["name", "channel", "unit", "epoch"],
            },
        )

    @staticmethod
    @pytest.fixture
    def ligolw(tmp_path) -> Path:
        """Return a `Path` with a LIGO_LW FrequencySeries array."""
        tmp = tmp_path / "test.xml"
        tmp.write_text(LIGO_LW_ARRAY)
        return tmp

    @pytest.mark.requires("lal", "igwn_ligolw")
    def test_read_ligolw(self, ligolw):
        """Test reading a `FrequencySeries` from ``LIGO_LW`` array."""
        array = FrequencySeries.read(ligolw, "PSD1")
        utils.assert_quantity_equal(
            array,
            [1, 2, 3, 4, 5] / units.Hz,
        )
        utils.assert_quantity_equal(
            array.frequencies,
            [10, 11, 12, 13, 14] * units.Hz,
        )
        assert numpy.isclose(array.epoch.gps, 1000000000)  # precision gah!
        assert array.unit == units.Hz ** -1

    @pytest.mark.requires("lal", "igwn_ligolw")
    def test_read_ligolw_params(self, ligolw):
        """Test reading a `FrequencySeries` from ``LIGO_LW`` array with params."""
        array = FrequencySeries.read(
            ligolw,
            channel="X1:TEST-CHANNEL_2",
        )
        assert list(array.value) == [10, 20, 30, 40, 50]
        assert array.epoch is None

    @pytest.mark.requires("lal", "igwn_ligolw")
    @pytest.mark.parametrize(("args", "match"), [
        # no name given, 'name' in error message
        pytest.param(
            [],
            "read: 'channel', 'epoch', 'f0', 'name'$",
            id="no-name",
        ),
        # name given, 'name' not in error message
        pytest.param(
            ("PSD2",),
            "read: 'channel', 'epoch', 'f0'$",
            id="name-only",
        ),
    ])
    def test_read_ligolw_error_multiple_array(self, args, match, ligolw):
        """Test reading a `FrequencySeries` from ``LIGO_LW`` with multiple <Array>."""
        # assert errors
        with pytest.raises(
            ValueError,
            match=match,
        ):  # multiple <Array> hits
            FrequencySeries.read(ligolw, *args)

    @pytest.mark.requires("lal", "igwn_ligolw")
    def test_read_ligolw_error_no_array(self, ligolw):
        """Test reading a `FrequencySeries` from ``LIGO_LW`` with no <Array>."""
        with pytest.raises(ValueError, match=r"^no <Array> elements found"):
            FrequencySeries.read(ligolw, "blah")

    @pytest.mark.requires("lal", "igwn_ligolw")
    def test_read_ligolw_error_no_match(self, ligolw):
        """Test reading a `FrequencySeries` from ``LIGO_LW`` with no matching Array."""
        with pytest.raises(
            ValueError,
            match="no <Array> elements found matching request",
        ):
            FrequencySeries.read(ligolw, epoch=0)

        with pytest.raises(
            ValueError,
            match="no <Array> elements found matching request",
        ):
            FrequencySeries.read(
                ligolw,
                "PSD1",
                f0=0,
            )

    @pytest.mark.requires("lal", "igwn_ligolw")
    def test_read_ligolw_error_no_param(self, ligolw):
        """Test reading a `FrequencySeries` from ``LIGO_LW`` with invalid param."""
        with pytest.raises(
            ValueError,
            match="no <Array> elements found matching request",
        ):
            FrequencySeries.read(
                ligolw,
                "PSD2",
                blah="blah",
            )

    @pytest.mark.requires("lal", "igwn_ligolw")
    def test_read_ligolw_error_dim(self, ligolw):
        """Test reading a `FrequencySeries` from ``LIGO_LW`` with invalid dim."""
        with pytest.raises(
            ValueError,
            match="cannot parse LIGO_LW Array with 3 dimensions in a Series",
        ):
            FrequencySeries.read(ligolw, epoch=1000000001)



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwpy.plot.text`."""

import pytest
from astropy.units import Unit
from matplotlib import (
    pyplot,
    rc_context,
)

from .. import text as plot_text


@pytest.mark.parametrize(("in_", "out", "texout"), [
    pytest.param(
        "test",
        "test",
        "test",
        id="str",
    ),
    pytest.param(
        4.0,
        "4.0",
        "4",
        id="float",
    ),
    pytest.param(
        8,
        "8",
        "8",
        id="int",
    ),
    pytest.param(
        Unit("m/Hz2"),
        "$\\mathrm{m\\,Hz^{-2}}$",
        "$\\mathrm{m\\,Hz^{-2}}$",
        id="unit",
    ),
])
def test_to_string(in_, out, texout):
    """Test `to_string()`."""
    with rc_context(rc={"text.usetex": False}):
        assert plot_text.to_string(in_) == out
    with rc_context(rc={"text.usetex": True}):
        assert plot_text.to_string(in_) == texout


@pytest.mark.parametrize(("unit", "label", "format_", "result"), [
    pytest.param(
        Unit("m"),
        None,
        "latex_inline_dimensional",
        r"Length [$\mathrm{m}$]",
        id="unit",
    ),
    pytest.param(
        Unit("m"),
        "My label",
        "latex_inline_dimensional",
        r"My label",
        id="label",
    ),
    pytest.param(
        Unit("m"),
        None,
        "console",
        "m",
        id="format",
    ),
])
def test_default_unit_label(unit, label, format_, result):
    """Test `default_unit_label()`."""
    fig = pyplot.figure()
    ax = fig.gca()
    if label:
        ax.set_xlabel(label)
    assert plot_text.default_unit_label(ax.xaxis, unit, format=format_) == result
    if not label:
        assert ax.xaxis.isDefault_label



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utility module to initialise Kerberos ticket-granting tickets.

This module provides a lazy-mans python version of the 'kinit'
command-line tool using the python-gssapi library.

See the documentation of the `kinit` function for example usage.
"""


from __future__ import annotations

import getpass
import logging
import os
import re
import subprocess
import sys
import warnings
from collections import OrderedDict
from pathlib import Path
from shutil import which
from typing import TYPE_CHECKING
from unittest import mock

if TYPE_CHECKING:
    from gssapi import (
        Credentials,
        Name,
    )

    from .utils import FileSystemPath

from ..utils.decorators import deprecated_function

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "kinit",
]

logger = logging.getLogger(__name__)

KLIST = which("klist") or "klist"

try:
    _IPYTHON = __IPYTHON__  # type: ignore[name-defined]
except NameError:
    _IPYTHON = False


# -- exceptions ---------------------------------

class KerberosError(RuntimeError):
    """Kerberos (krb5) operation failed."""



# -- utilities ----------------------------------

def _validate_keytab(
    username: str | None = None,
    realm: str | None = None,
    keytab: FileSystemPath | None = None,
) -> tuple[str | None, str | None]:
    """Get the default keytab and check that a file exists.

    This method reformats the ``username`` into a full principal name, as best
    it can, using the given ``realm`` or one pulled out of the ``keytab``.

    Parameters
    ----------
    username : `str`
        The username for the Kerberos principal (possibly including a realm,
        as given by the user to `kinit()`).

    realm : `str`
        The realm of the Kerberos principal
        (as given by the user to `kinit()`).

    keytab : `str`, `pathlib.Path`
        The path of the Kerberos keytab to use (as given by the user to `kinit()`).

    Returns
    -------
    principal : `str`
        The best-formatted full principal name.

    keytab : `str`
        The keytab path.

    Raises
    ------
    gwpy.io.kerberos.KerberosError
        If the ``username`` and/or ``realm`` conflict with that discovered
        in the ``keytab`` (if given).

    gssapi.exceptions.GSSError
        If ``keytab`` is given but a principal cannot be parsed from it.
    """
    import gssapi

    # if the user gave the keytab, use it
    if keytab:
        strict = True
    # otherwise try the environment, but be lenient
    else:
        keytab = os.environ.get("KRB5_KTNAME", None)
        strict = False

    # no keytab, use the username we were given
    if keytab is None:
        return username, None

    keytab = str(Path(keytab).expanduser())
    try:  # get principal from default keytab
        principal = _keytab_principal(keytab)
        puser, prealm = principal.split("@", 1)
    except (
        gssapi.exceptions.GSSError,
        ValueError,
    ):
        if strict:
            raise
        return username, None

    # user didn't specify a name or a realm, use what we got from the keytab
    if not username and not realm:
        return principal, keytab

    # user _did_ specify username or realm, so if they match the keytab,
    # then we can use it
    try:
        user, urealm = username.split("@", 1)  # type: ignore[union-attr]
    except (
        AttributeError,  # username is None
        ValueError,  # username doesn't contain an '@'
    ):
        user = username
        urealm = realm
    if (
        (user is None or user == puser)
        and (urealm is None or urealm == prealm)
    ):
        return principal, keytab

    if strict:
        message = (
            f"principal '{principal}' from keytab ('{keytab}') doesn't match "
            f"username '{username}'"
        )
        if realm:
            message += f", or realm '{realm}'"
        raise KerberosError(message)

    return username, None


def _check_interactive(
    username: str | None,
    password: str | None,
    keytab: FileSystemPath | None,
) -> None:
    """Check that we can prompt for necessary information.

    This function raises an exception if we need to prompt for
    information but we are not in an interactive session.
    """
    _prompt_username = username is None
    _prompt_password = not keytab and password is None
    if (
        not sys.stdout.isatty()
        and not _IPYTHON
        and (_prompt_username or _prompt_password)
    ):
        msg = (
            "cannot generate Kerberos ticket in a non-interactive session, "
            "please manually create a ticket, or consider using a keytab file"
        )
        raise KerberosError(msg)


def _get_principal(
    username: str | None,
    realm: str | None,
) -> Name:
    """Prompt for and canonicalise the principal name."""
    import gssapi

    if username is None:
        username = input(
            f"Kerberos principal (user@{realm or 'REALM'}): ",
        )
    if "@" not in username and realm:
        username = f"{username}@{realm}"
    principal = gssapi.Name(
        base=username,
        name_type=gssapi.NameType.kerberos_principal,
    )
    try:
        # applies default realm if not given
        return principal.canonicalize(
            gssapi.MechType.kerberos,
        )
    except gssapi.exceptions.GSSError as exc:
        msg = "failed to canonicalize Kerberos principal name, please specify `realm`"
        raise KerberosError(msg) from exc


def _acquire_keytab(
    principal: Name,
    keytab: FileSystemPath,
    ccache: FileSystemPath | None = None,
    lifetime: int | None = None,
) -> Credentials:
    """Acquire a Kerberos TGT using a keytab."""
    import gssapi
    keytab = os.fspath(keytab)
    store: dict[bytes | str, bytes | str] = {
        "client_keytab": keytab,
    }
    if ccache:
        store["ccache"] = os.fspath(ccache)
    with mock.patch.dict("os.environ", {"KRB5_KTNAME": keytab}):
        creds = gssapi.Credentials(
            name=principal,
            store=store,
            usage="initiate",
            lifetime=lifetime,
        )
    creds.inquire()
    return creds


def _acquire_password(
    principal: Name,
    password: str,
    ccache: FileSystemPath | None = None,
    lifetime: int | None = None,
) -> Credentials:
    """Acquire a Kerberos TGT using principal/password."""
    import gssapi
    raw_creds = gssapi.raw.acquire_cred_with_password(
        principal,
        password.encode("utf-8"),
        lifetime=lifetime,
        usage="initiate",
    )
    creds = gssapi.Credentials(base=raw_creds.creds)
    creds.inquire()
    creds.store(
        store={"ccache": os.fspath(ccache)} if ccache else None,
        usage="initiate",
        overwrite=True,
    )
    return creds


def _keytab_principal(
    keytab: FileSystemPath,
) -> str:
    """Return the principal assocated with a Kerberos keytab file."""
    import gssapi
    with mock.patch.dict("os.environ", {"KRB5_KTNAME": os.fspath(keytab)}):
        return str(gssapi.Credentials(usage="accept").name)


def kinit(
    username: str | None = None,
    password: str | None = None,
    realm: str | None = None,
    keytab: FileSystemPath | None = None,
    ccache: FileSystemPath | None = None,
    lifetime: int | None = None,
    krb5ccname: str | None = None,
    *,
    verbose: bool | None = None,
) -> Credentials:
    """Initialise a Kerberos ticket-granting ticket (TGT).

    Parameters
    ----------
    username : `str`, optional
        Name principal for Kerberos credential, will be prompted for
        if not given.

    password : `str`, optional
        Cleartext password of user for given realm, will be prompted for
        if not given.

        .. warning::

            Passing passwords in plain text presents a security risk, please
            consider using a Kerberos keytab file to store credentials.

    realm : `str`, optional
        Name of realm to authenticate against if not given as part of
        ``username``.
        Defaults to ``'default_realm'``; see ``man krb5.conf(5)``.

    keytab : `str`, `~pathlib.Path`, optional
        Path to keytab file. If not given this will be read from the
        ``KRB5_KTNAME`` environment variable. See notes for more details.

    ccache : `str`, `~pathlib.Path`, optional
        Path to Kerberos credentials cache.

    lifetime : `int`, optional
        Desired liftime of the Kerberos credential (may not be respected
        by the underlying GSSAPI implementation); pass `None` to use
        the maximum permitted liftime (default).

        This is currently not respected by MIT Kerberos (the most common
        GSSAPI implementation).

    verbose : `bool`, optional
        DEPRECATED. This argument does nothing, instead of using ``verbose=True``
        use logging to control output.

    krb5ccname : `str`, optional
        DEPRECATED. This argument has been renamed to ``ccache``.

    Returns
    -------
    creds : `gssapi.Credentials`
        The acquired Kerberos credentials.

    Notes
    -----
    If a keytab is given, or is read from the ``KRB5_KTNAME`` environment
    variable, this will be used to guess the principal, if it
    contains only a single credential.

    Examples
    --------
    Example 1: standard user input, with password prompt:

    >>> kinit('albert.einstein')
    Password for albert.einstein@LIGO.ORG:
    Kerberos ticket generated for albert.einstein@LIGO.ORG

    Example 2: extract username and realm from keytab, and use that
    in authentication:

    >>> kinit(keytab='~/.kerberos/ligo.org.keytab', verbose=True)
    Kerberos ticket generated for albert.einstein@LIGO.ORG
    """
    try:
        import gssapi
    except ImportError as exc:
        msg = (
            "cannot generate Kerberos credentials without python-gssapi, "
            "or run `kinit` from your terminal manually."
        )
        raise ImportError(msg) from exc

    # handle deprecated keywords
    if krb5ccname:
        warnings.warn(
            f"The `krb5ccname` keyword for {__name__}.kinit was renamed "
            "to `ccache`, and will stop working in a future release.",
            DeprecationWarning,
            stacklevel=2,
        )
        if ccache is None:
            ccache = krb5ccname
    if verbose is not None:
        warnings.warn(
            f"The `verbose` keyword for {__name__}.kinit is deprecated, "
            "and will stop working in a future release. "
            "Please use logging to control output.",
            DeprecationWarning,
            stacklevel=2,
        )

    # get keytab and check we can use it (username in keytab)
    try:
        username, keytab = _validate_keytab(username, realm, keytab)
    except gssapi.exceptions.GSSError as exc:
        msg = f"Kerberos keytab '{keytab}' is invalid, see traceback for full details"
        raise KerberosError(msg) from exc

    # refuse to prompt if we can't get an answer
    # note: jupyter streams are not recognised as interactive
    #       (isatty() returns False) so we have a special case here
    _check_interactive(
        username,
        password,
        keytab,
    )

    # get username
    principal = _get_principal(
        username,
        realm,
    )

    # get password
    if not keytab and password is None:
        password = getpass.getpass(prompt=f"Password for {principal}: ")

    # generate credential
    try:
        if keytab:
            creds = _acquire_keytab(
                principal,
                keytab,
                ccache=ccache,
                lifetime=lifetime,
            )
        else:
            creds = _acquire_password(
                principal,
                str(password),
                ccache=ccache,
                lifetime=lifetime,
            )
    except gssapi.exceptions.GSSError as exc:
        msg = (
            f"failed to generate Kerberos TGT for {principal}, "
            "see traceback for full details"
        )
        raise KerberosError(msg) from exc

    logger.info(
        "Kerberos ticket acquired for %s (%d seconds remaining)",
        creds.name,
        creds.lifetime,
    )

    return creds


# -- deprecated ---------------------------------

@deprecated_function
def parse_keytab(keytab: str) -> list[tuple[str, str, int]]:  # pragma: no cover
    """Read the contents of a KRB5 keytab file, returning a list of credentials.

    Parameters
    ----------
    keytab : `str`
        path to keytab file

    Returns
    -------
    creds : `list` of `tuple`
        the (unique) list of `(username, realm, kvno)` as read from the
        keytab file

    Examples
    --------
    >>> from gwpy.io.kerberos import parse_keytab
    >>> print(parse_keytab("creds.keytab"))
    [('albert.einstein', 'LIGO.ORG', 1)]
    """
    try:
        out = subprocess.check_output(
            [KLIST, "-k", keytab],
            stderr=subprocess.PIPE,
        )
    except OSError as exc:
        msg = "failed to locate klist, cannot read keytab"
        raise KerberosError(msg) from exc
    except subprocess.CalledProcessError as exc:
        msg = f"cannot read keytab '{keytab}'"
        raise KerberosError(msg) from exc
    principals: list[tuple[str, str, int]] = []
    line: str | bytes
    for line in out.splitlines():
        if isinstance(line, bytes):
            line = line.decode("utf-8")  # noqa: PLW2901
        try:
            kvno, principal, = re.split(r"\s+", line.strip(" "), maxsplit=1)
        except ValueError:
            continue
        else:
            if not kvno.isdigit():
                continue
            principals.append((*principal.split("@"), int(kvno)))
    # return unique, ordered list
    return list(OrderedDict.fromkeys(principals))



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read SegmentLists from seg-wizard format ASCII files."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

from ...io.registry import default_registry
from ...io.utils import with_open
from ...time import LIGOTimeGPS
from .. import (
    Segment,
    SegmentList,
)

if TYPE_CHECKING:
    from typing import (
        IO,
        TextIO,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

_FLOAT_PAT = r"([\d.+-eE]+)"
# simple two-column (gpsstart, gpsend)
TWO_COL_REGEX = re.compile(
    rf"\A\s*{_FLOAT_PAT}\s+{_FLOAT_PAT}\s*\Z",
)
# three column (gpsstart, gpsend, duration)
THREE_COL_REGEX = re.compile(
    rf"\A\s*{_FLOAT_PAT}\s+{_FLOAT_PAT}\s+{_FLOAT_PAT}\s*\Z",
)
# four column (index, gpsstart, gpsend, duration)
FOUR_COL_REGEX = re.compile(
    rf"\A\s*([\d]+)\s+{_FLOAT_PAT}\s+{_FLOAT_PAT}\s+{_FLOAT_PAT}\s*\Z",
)


# -- identify ------------------------

def _is_segwizard(
    origin: str,
    filepath: str,
    fileobj: IO,
    *args,
    **kwargs,
):
    """Return `True` if the given file looks like a segwizard file.

    When introspecting a file contents, this method only recognises the
    four-column seg/start/stop/duration format written by
    `igwn_segments.utils.tosegwizard`.
    """
    if fileobj is not None:
        pos = fileobj.seek(0)
        try:
            header = fileobj.readline()
        finally:
            fileobj.seek(pos)
        cols = header.strip().split()
        return cols == [b"#", b"seg", b"start", b"stop", b"duration"]
    if filepath is not None:
        return filepath.endswith((".dat", ".txt"))
    return False


# -- read ----------------------------

@with_open
def from_segwizard(
    source: IO,
    gpstype: type = LIGOTimeGPS,
    strict: bool = True,
) -> SegmentList:
    """Read segments from a segwizard format file into a `SegmentList`.

    Parameters
    ----------
    source : `file`, `str`, `pathlib.Path`
        An open file, or file path, from which to read.

    gpstype : `type`, optional
        The numeric type to which to cast times (from `str`) when reading.

    strict : `bool`, optional
        Check that recorded duration matches ``end-start`` for all segments;
        only used when reading from a 3+-column file.

    Returns
    -------
    segments : `~gwpy.segments.SegmentList`
        The list of segments as parsed from the file.

    Notes
    -----
    This method is adapted from original code written by Kipp Cannon and
    distributed under GPLv3.
    """
    # read file object
    out = SegmentList()
    fmt_pat = None
    for line in source:
        if isinstance(line, bytes):
            line = line.decode("utf-8")
        if line.startswith(("#", ";")):  # comment
            continue
        # determine line format
        if fmt_pat is None:
            fmt_pat = _line_format(line)
        # parse line
        tokens, = fmt_pat.findall(line)
        out.append(_format_segment(
            tokens[-3:],
            gpstype=gpstype,
            strict=strict,
        ))
    return out


def _line_format(
    line: str,
) -> re.Pattern:
    """Determine the column format pattern for a line in an ASCII segment file."""
    for pat in (
        FOUR_COL_REGEX,
        THREE_COL_REGEX,
        TWO_COL_REGEX,
    ):
        if pat.match(line):
            return pat
    msg = f"unable to parse segment from line '{line}'"
    raise ValueError(msg)


def _format_segment(
    tokens: list[str],
    strict: bool = True,
    gpstype: type = LIGOTimeGPS,
) -> Segment:
    """Format a list of tokens parsed from an ASCII file into a segment."""
    try:
        start, end, dur = tokens
    except ValueError:  # two-columns
        return Segment(*map(gpstype, tokens))
    seg = Segment(gpstype(start), gpstype(end))
    if strict and abs(seg) != float(dur):
        msg = f"segment {seg} has incorrect duration {dur}"
        raise ValueError(msg)
    return seg


# -- write ---------------------------

@with_open(mode="w", pos=1)
def to_segwizard(
    segs: SegmentList,
    target: TextIO,
    header: bool = True,
    coltype: type = LIGOTimeGPS,
):
    """Write the given `SegmentList` to a file in SegWizard format.

    Parameters
    ----------
    segs : :class:`~gwpy.segments.SegmentList`
        The list of segments to write.

    target : `io.TextIOBase`
        A file open for writing in text mode.

    header : `bool`, optional
        Print a column header into the file, default: `True`.

    coltype : `type`, optional
        The numerical type in which to cast times before printing.

    Notes
    -----
    This method is adapted from original code written by Kipp Cannon and
    distributed under GPLv3.
    """
    # write file object
    if header:
        print("# seg\tstart\tstop\tduration", file=target)
    for i, seg in enumerate(segs):
        a = coltype(seg[0])
        b = coltype(seg[1])
        c = float(b - a)
        print(
            "\t".join(map(str, (i, a, b, c))),
            file=target,
        )


# -- register ------------------------

default_registry.register_reader("segwizard", SegmentList, from_segwizard)
default_registry.register_writer("segwizard", SegmentList, to_segwizard)
default_registry.register_identifier(
    "segwizard",
    SegmentList,
    _is_segwizard,
)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read and write LIGO Channel List Files.

The Channel List File (CLF) is a schema of the INI configuration file format
designed to hold lists of interferometer data channels for bulk processing.

Each CLF file should contain one or more `[groups]` each containing, at
least, an option named `channels` with a newline-delimited list of channel
names, e.g.

.. code-block:: ini

    [my channels]
    channels =
        X1:CHANNEL-1
        X1:CHANNEL-2

Each channel line can also optionally contain a sampling rate number, e.g.

.. code-block:: ini

    [my channels]
    channels =
       X1:CHANNEL-1
       X1:CHANNEL-2
       X1:CHANNEL-3 2048

Other recommended options to include with a channel group include

  - `frametype`: the GWF type for files containing data for these channels
  - `flow`: the lower-frequency bound for processing these channels
  - `fhigh`: the upper-frequency bound for processing these channels
  - `qhigh`: the upper bound on sine-Gaussian Q for processing these channels

For example,

.. code-block:: ini

    [ALS]
    flow = 32
    qhigh = 100
    frametype = L1_R
    channels =
        L1:ALS-X_ARM_IN1_DQ 2048
        L1:ALS-Y_ARM_IN1_DQ

"""

from __future__ import annotations

import configparser
import re
from typing import TYPE_CHECKING

from numpy import inf

from ...io.registry import identify_factory
from ...io.utils import (
    file_list,
    with_open,
)
from .. import (
    Channel,
    ChannelList,
)

if TYPE_CHECKING:
    from pathlib import Path
    from typing import IO

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

CHANNEL_DEFINITION = re.compile(
    r"(?P<name>[a-zA-Z0-9:_-]+)"
    r"(?:\s+(?P<sample_rate>[0-9.]+))?"
    r"(?:\s+(?P<safe>(safe|unsafe|unsafeabove2kHz|unknown)))?"
    r"(?:\s+(?P<fidelity>(clean|flat|glitchy|unknown)))?",
)


# -- read ----------------------------

def read_channel_list_file(*source: str | Path | IO) -> ChannelList:
    """Read a `~gwpy.detector.ChannelList` from a Channel List File."""
    # read file(s)
    config = configparser.ConfigParser()
    files = file_list(source)
    success_ = config.read(files)
    if failed := set(files) - set(success_):
        msg = f"Failed to read '{failed.pop()}'"
        raise OSError(msg)

    # create channel list
    out = ChannelList()
    out.source = source
    append = out.append

    # loop over all groups and channels
    for group in config.sections():
        params = dict(config.items(group))

        # parse flow and fhigh as 'frange'
        if "flow" in params or "fhigh" in params:
            low = params.pop("flow", 0)
            high = params.pop("fhigh", inf)
            if isinstance(high, str) and high.lower() == "nyquist":
                high = inf
            frange = float(low), float(high)
        else:
            frange = None

        for channel in params.pop("channels").strip("\n").split("\n"):
            append(_read_channel(channel, group, frange, **params))

    return out


def _read_channel(
    value: str,
    group: str,
    frange: tuple[float, float] | None,
    **params,
) -> Channel:
    """Parse a single channel."""
    if not (match := CHANNEL_DEFINITION.match(value)):
        msg = f"Cannot parse '{value}' as channel list entry"
        raise ValueError(msg)

    # remove Nones from match
    inlineattrs = {k: v for k, v in match.groupdict().items() if v is not None}
    inlineattrs.setdefault("fidelity", "clean")

    # parse safe
    safe = inlineattrs.get("safe", "safe").lower() != "unsafe"

    # create channel and copy group params
    channel = Channel(
        inlineattrs.pop("name"),
        frequency_range=frange,
        safe=safe,
        sample_rate=inlineattrs.pop("sample_rate"),
    )
    channel.params = params | inlineattrs
    channel.group = group

    # extract those params for which the Channel has an attribute
    for key in ["frametype"]:
        setattr(channel, key, channel.params.pop(key, None))

    return channel


# -- write ---------------------------

@with_open(mode="w", pos=1)
def write_channel_list_file(
    channels: ChannelList,
    fobj: IO,
) -> None:
    """Write a `~gwpy.detector.ChannelList` to a INI-format channel list file."""
    # create the configparser
    out = configparser.ConfigParser()
    for channel in channels:
        _write_channel_ini(out, channel)

    # write to file
    out.write(fobj)


def _write_channel_ini(
    ini: configparser.ConfigParser,
    channel: Channel,
) -> None:
    """Write a `Channel` to a `~configparser.ConfigParser`."""
    group = channel.group
    if not ini.has_section(group):
        ini.add_section(group)
    for param, value in channel.params.items():
        ini.set(group, param, value)
    if channel.sample_rate is not None:
        entry = f"{channel} {channel.sample_rate.to('Hz').value}"
    else:
        entry = str(channel)
    entry += f" {channel.params.get('safe', 'safe')}"
    entry += f" {channel.params.get('fidelity', 'clean')}"
    try:
        clist = ini.get(group, "channels")
    except configparser.NoOptionError:
        ini.set(group, "channels", f"\n{entry}")
    else:
        ini.set(group, "channels", clist + f"\n{entry}")


# -- Unified I/O registration --------

ChannelList.read.registry.register_reader(
    "ini",
    ChannelList,
    read_channel_list_file,
)
ChannelList.read.registry.register_identifier(
    "ini",
    ChannelList,
    identify_factory(".ini", ".clf"),
)
ChannelList.write.registry.register_writer(
    "ini",
    ChannelList,
    write_channel_list_file,
)



# Copyright (c) 2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.io.gwf.backend`."""

import os
from unittest import mock

import pytest

from .. import gwf as io_gwf

BACKEND_LIBRARY = {
    "frameCPP": "LDAStools.frameCPP",
    "FrameL": "framel",
    "LALFrame": "lalframe",
}
# need at least one backend (LALFrame is most widely available)
ANY_BACKEND = pytest.mark.requires(BACKEND_LIBRARY["LALFrame"])


@mock.patch.dict("os.environ")
@ANY_BACKEND
def test_get_backend():
    """Test that :func:`get_backend` returns a sensible value."""
    os.environ.pop("GWPY_FRAME_LIBRARY", None)
    assert io_gwf.get_backend() in io_gwf.BACKENDS


@mock.patch.dict("os.environ")
@pytest.mark.parametrize("backend", [
    pytest.param(backend, marks=pytest.mark.requires(BACKEND_LIBRARY[backend]))
    for backend in io_gwf.BACKENDS
])
def test_get_backend_environ(backend):
    """Test that :func:`get_backend` reads the environment poroperly."""
    os.environ["GWPY_FRAME_LIBRARY"] = backend
    assert io_gwf.get_backend() == backend


@mock.patch.dict("os.environ")
@ANY_BACKEND
def test_get_backend_environ_bad():
    """Test that `get_backend_environ` returns a valid value despiate a bad env."""
    os.environ["GWPY_FRAME_LIBRARY"] = "blahblahblah"
    assert io_gwf.get_backend() in io_gwf.BACKENDS


def test_get_backend_function():
    """Test that `get_backend_function` works."""
    io_framecpp = pytest.importorskip("gwpy.io.gwf.framecpp")
    func = io_gwf.get_backend_function("create_frame", backend="frameCPP")
    assert func is io_framecpp.create_frame


def test_get_backend_function_badbackend():
    """Test that `get_backend_function` formats errors correctly."""
    with pytest.raises(
        ImportError,
        match=r"No module named 'gwpy.io.gwf.notimplemented'",
    ):
        io_gwf.get_backend_function("notimplemented", backend="notimplemented")


@ANY_BACKEND
def test_get_backend_function_notimplemented():
    """Test that `get_backend_function` formats errors correctly."""
    with pytest.raises(
        NotImplementedError,
        match="'notimplemented'",
    ):
        io_gwf.get_backend_function("notimplemented")



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for `gwpy.table`.

Note that tests of the I/O integrations are separately maintained
in the ``test_io_{format}.py`` modules alongside this one.
"""

from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Generic,
    TypeVar,
)

import pytest
from astropy import units
from numpy.testing import assert_array_equal

from ...segments import (
    Segment,
    SegmentList,
)
from ...testing import utils
from ...time import LIGOTimeGPS
from ...timeseries import (
    TimeSeries,
    TimeSeriesDict,
)
from .. import (
    EventTable,
    Table,
    filters,
)
from .utils import random_table

if TYPE_CHECKING:
    import numpy

TableType = TypeVar("TableType", bound=Table)
EventTableType = TypeVar("EventTableType", bound=EventTable)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

TEST_DATA_PATH = Path(__file__).parent / "data"
TEST_XML_FILE = TEST_DATA_PATH / "H1-LDAS_STRAIN-968654552-10.xml.gz"
TEST_OMEGA_FILE = TEST_DATA_PATH / "omega.txt"


# -- Test gwpy.table.Table (astropy.table.Table)

class TestTable(Generic[TableType]):
    """Tests modifications/extensions to `astropy.table.Table`."""

    TABLE: type[TableType] = Table  # type: ignore[assignment]

    @classmethod
    def create(cls, n, names, dtypes=None) -> TableType:
        """Create a random table of the given length, names and dtypes."""
        return random_table(
            names=names,
            length=n,
            dtypes=dtypes,
            table_class=cls.TABLE,
        )

    @pytest.fixture
    @classmethod
    def table(cls) -> TableType:
        """Return a random table with some typical columns."""
        return cls.create(100, ["time", "snr", "frequency"])

    @pytest.fixture
    @classmethod
    def emptytable(cls) -> TableType:
        """Return an empty table with some typical columns."""
        return cls.create(0, ["time", "snr", "frequency"])

    @pytest.fixture
    @classmethod
    def clustertable(cls) -> TableType:
        """Return a small table suitable for clustering tests."""
        return cls.TABLE(
            data=[
                [11, 1, 1, 10, 1, 1, 9],
                [0.0, 1.9, 1.95, 2.0, 2.05, 2.1, 4.0],
            ],
            names=["amplitude", "time"],
        )


class TestEventTable(TestTable[EventTableType]):
    """Tests for `EventTable`."""

    TABLE: type[EventTableType] = EventTable  # type: ignore[assignment]

    def test_get_time_column(self, table: EventTableType):
        """Test that `_get_time_column` works on name."""
        assert table._get_time_column() == "time"

    def test_get_time_column_case(self, table: EventTableType):
        """Test that `_get_time_column` works on name case-insensitively."""
        table.rename_column("time", "TiMe")
        assert table._get_time_column() == "TiMe"

    def test_get_time_column_gps_type(self):
        """Test that `_get_time_column` works on dtype."""
        # check that single GPS column can be identified
        t = self.create(1, ("a", "b"), dtypes=(float, LIGOTimeGPS))
        assert t._get_time_column() == "b"

    def test_get_time_column_error_no_match(self):
        """Test `_get_time_column` error handling when it fails to find a match."""
        t = self.create(1, ("a",))
        with pytest.raises(
            ValueError,
            match=" no columns named 'time'",
        ):
            t._get_time_column()

    def test_get_time_column_error_multiple_match(self):
        """Test `_get_time_column` error handling when it finds multiple matches."""
        # check that two GPS columns causes issues
        t = self.create(
            10,
            ("a", "b", "c"),
            dtypes=(float, LIGOTimeGPS, LIGOTimeGPS),
        )
        with pytest.raises(
            ValueError,
            match=" multiple columns named 'time'",
        ):
            t._get_time_column()

    def test_get_time_column_error_empty(self):
        """Check that `_get_time_column` errors properly on an empty table."""
        t = self.create(0, ("a",))
        with pytest.raises(
            ValueError,
            match="cannot identify time column for table",
        ):
            t._get_time_column()

    def test_filter(self, table: EventTableType):
        """Test that `EventTable.filter` works with a simple filter statement."""
        # check simple filter
        lowf = table.filter("frequency < 100")
        assert isinstance(lowf, type(table))
        assert len(lowf) == 11
        assert lowf["frequency"].max() == pytest.approx(99.41111793264079)

    def test_filter_empty(self, table: EventTableType):
        """Test that `EventTable.filter` works with an empty table."""
        assert len(table.filter("snr>5", "snr<=5")) == 0

    def test_filter_chaining(self, table: EventTableType):
        """Test that chaining filters works with `EventTable.filter`."""
        loud = table.filter("snr > 100")
        lowf = table.filter("frequency < 100")
        lowfloud = table.filter("frequency < 100", "snr > 100")
        brute = type(table)(
            rows=[tuple(row) for row in lowf if row in loud],
            names=table.dtype.names,
        )
        utils.assert_table_equal(brute, lowfloud)

    def test_filter_range(self, table: EventTableType):
        """Test that `EventTable.filter` works with a range statement."""
        # check double-ended filter
        midf = table.filter("100 < frequency < 1000")
        utils.assert_table_equal(
            midf,
            table.filter("frequency > 100").filter("frequency < 1000"),
        )

    def test_filter_function(self, table: EventTableType):
        """Test that `EventTable.filter` works with a filter function."""
        def my_filter(column, threshold):
            return column < threshold

        lowf = table.filter(("frequency", my_filter, 100))
        assert len(lowf) == 11

    def test_filter_function_multiple(self, table: EventTableType):
        """Test that `EventTable.filter` works with a multi-column filter."""
        def my_filter(table, threshold) -> numpy.ndarray:
            return table["snr"] * table["frequency"] > threshold

        filtered = table.filter((("snr", "frequency"), my_filter, 100000))
        assert len(filtered) == 67

    def test_filter_in_segmentlist(self, table: EventTableType):
        """Test `EventTable.filter` with `in_segmentlist`."""
        # check filtering on segments works
        segs = SegmentList([Segment(100, 200), Segment(400, 500)])
        inseg = table.filter(("time", filters.in_segmentlist, segs))
        brute = type(table)(
            rows=[tuple(row) for row in table if row["time"] in segs],
            names=table.colnames,
        )
        utils.assert_table_equal(inseg, brute)

    def test_filter_in_segmentlist_empty(self, table: EventTableType):
        """Test `EventTable.filter` with `in_segmentlist` and an empty table."""
        # check empty segmentlist is handled well
        utils.assert_table_equal(
            table.filter(("time", filters.in_segmentlist, SegmentList())),
            type(table)(names=table.colnames),
        )

    def test_filter_not_in_segmentlist(self, table: EventTableType):
        """Test `EventTable.filter` with `not_in_segmentlist`."""
        segs = SegmentList([Segment(100, 200), Segment(400, 500)])
        notsegs = SegmentList([Segment(0, 1000)]) - segs
        inseg = table.filter(("time", filters.in_segmentlist, segs))
        utils.assert_table_equal(
            inseg, table.filter(("time", filters.not_in_segmentlist, notsegs)),
        )
        utils.assert_table_equal(
            table,
            table.filter(("time", filters.not_in_segmentlist, SegmentList())),
        )

    def test_event_rates(self, table: EventTableType):
        """Test :meth:`gwpy.table.EventTable.event_rate`."""
        rate = table.event_rate(1)
        assert isinstance(rate, TimeSeries)
        assert rate.sample_rate == 1 * units.Hz

    @pytest.mark.requires("lal")
    def test_event_rates_gpsobject(self, table: EventTableType):
        """Test that `EventTable.event_rate` can handle object dtypes."""
        rate = table.event_rate(1)

        from lal import LIGOTimeGPS as LalGps
        lgps = list(map(LalGps, table["time"]))
        t2 = type(table)(data=[lgps], names=["time"])
        rate2 = t2.event_rate(1, start=table["time"].min())

        utils.assert_quantity_sub_equal(rate, rate2, exclude=["epoch", "x0"])

    def test_event_rates_start_end(self):
        """Test that `EventTable.event_rate` works without time column.

        If and only if start/end are both given.
        """
        t2 = self.create(10, names=["a", "b"])
        with pytest.raises(
            ValueError,
            match="please give `timecolumn` keyword",
        ):
            t2.event_rate(1)
        with pytest.raises(ValueError, match="cannot identify time column"):
            t2.event_rate(1, start=0)
        with pytest.raises(ValueError, match="cannot identify time column"):
            t2.event_rate(1, end=1)
        t2.event_rate(1, start=0, end=10)

    def test_binned_event_rates(self, table: EventTableType):
        """Test :meth:`binned_event_rates`."""
        rates = table.binned_event_rates(
            100,
            "snr",
            [10, 100],
            timecolumn="time",
        )
        assert isinstance(rates, TimeSeriesDict)
        assert list(rates.keys()), [10, 100]
        assert rates[10].max() == 0.16 * units.Hz   # type: ignore[index]
        assert rates[10].name == "snr >= 10"        # type: ignore[index]
        assert rates[100].max() == 0.15 * units.Hz  # type: ignore[index]
        assert rates[100].name == "snr >= 100"      # type: ignore[index]
        table.binned_event_rates(100, "snr", [10, 100], operator="in")
        table.binned_event_rates(100, "snr", [(0, 10), (10, 100)])

    def test_binned_event_rates_start_end(self):
        """Test that `EventTable.binned_event_rates` works without time column.

        If and only if start/end are both given.
        """
        table = self.create(0, names=["a", "b"])
        with pytest.raises(
            ValueError,
            match="please give `timecolumn` keyword",
        ):
            table.binned_event_rates(1, "a", (10, 100))
        with pytest.raises(ValueError, match="cannot identify time column"):
            table.binned_event_rates(1, "a", (10, 100), start=0)
        with pytest.raises(ValueError, match="cannot identify time column"):
            table.binned_event_rates(1, "a", (10, 100), end=1)
        table.binned_event_rates(1, "a", (10, 100), start=0, end=10)

    def test_scatter(self, table: EventTableType):
        """Test `EventTable.scatter`."""
        plot = table.scatter("time", "frequency", color="snr")
        plot.save(BytesIO(), format="png")
        plot.close()

    def test_scatter_default_label(self, table: EventTableType):
        """Test `EventTable.scatter` default axis labeling.

        Checks regression against https://gitlab.com/gwpy/gwpy/-/issues/1844.
        """
        plot = table.scatter("time", "frequency", ylabel="My label")
        try:
            assert plot.gca().get_xlabel() == "time"
            assert plot.gca().get_ylabel() == "My label"
        finally:
            plot.close()

    def test_hist(self, table: EventTableType):
        """Test `EventTable.hist`."""
        plot = table.hist("snr")
        assert len(plot.gca().patches) == 10
        plot.save(BytesIO(), format="png")
        plot.close()

    def test_get_column(self, table: EventTableType):
        """Test `EventTable.get_column`."""
        utils.assert_array_equal(table.get_column("snr"), table["snr"])

    def test_cluster(self, clustertable: EventTableType):
        """Test `EventTable.cluster`."""
        # check that the central data points are all clustered away,
        # the original table is unchanged, and all points return their
        # intended values
        t = clustertable.cluster("time", "amplitude", 0.6)
        assert len(t) == 3
        assert len(clustertable) == 7
        assert_array_equal(t["amplitude"], [11, 10, 9])
        assert_array_equal(t["time"], [0.0, 2.0, 4.0])

    def test_single_point_cluster(self, clustertable: EventTableType):
        """Test `EventTable.cluster` with a large window."""
        # check that a large cluster window returns at least one data point
        t = clustertable.cluster("time", "amplitude", 10)
        assert len(t) == 1
        assert_array_equal(t["amplitude"], [11])
        assert_array_equal(t["time"], [0.0])

    def test_cluster_window(self, clustertable: EventTableType):
        """Test `EventTable.cluster` window parameter error handling."""
        # check that a non-positive window throws an appropriate ValueError
        with pytest.raises(
            ValueError,
            match=r"^Window must be a positive value$",
        ):
            clustertable.cluster("time", "amplitude", 0)

    def test_cluster_multiple(self, clustertable: EventTableType):
        """Test `EventTable.cluster` multiple calls consistency."""
        # check that after clustering a table, clustering the table a
        # second time with the same parameters returns the same result
        t_clustered = clustertable.cluster("time", "amplitude", 0.6)
        utils.assert_table_equal(
            t_clustered,
            t_clustered.cluster("time", "amplitude", 0.6),
        )

    def test_cluster_empty(self, emptytable: EventTableType):
        """Test `EventTable.cluster` with an empty table."""
        t = emptytable.cluster("time", "amplitude", 0.6)
        utils.assert_table_equal(t, emptytable)



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

from collections.abc import Iterator
from typing import (
    Any,
    overload,
)

class availability:  # noqa: N801
    data: segment_list_type

class availability_list_type(list[availability]): ...  # noqa: N801

class buffer:  # noqa: N801
    channel: channel
    channel_type: int
    data: Any  # numpy array-like
    data_type: int
    gps_seconds: int
    gps_nanoseconds: int
    length: int
    name: str
    sample_rate: float
    signal_gain: float
    signal_slope: float
    signal_offset: float
    signal_units: str

    def release(self) -> None: ...

class channel:  # noqa: N801
    # Class constants
    MIN_SAMPLE_RATE: float
    MAX_SAMPLE_RATE: float
    DEFAULT_CHANNEL_MASK: int
    DEFAULT_DATA_MASK: int

    # Instance attributes accessed in GWpy
    name: str
    sample_rate: float
    signal_units: str | None
    data_type: int
    channel_type: int

    # Static methods
    @staticmethod
    def channel_type_to_string(channel_type: int) -> str: ...

    @staticmethod
    def data_type_to_string(data_type: int) -> str: ...

class connection:  # noqa: N801
    def __init__(
        self,
        hostname: str,
        port: int = ...,
    ) -> None: ...

    def open(self, start_time: int, end_time: int) -> None: ...

    def close(self) -> None: ...

    def count_channels(
        self,
        channel_glob: str = ...,
        channel_type_mask: int = ...,
        data_type_mask: int = ...,
        min_sample_rate: float = ...,
        max_sample_rate: float = ...,
    ) -> int: ...

    def current_epoch(self) -> segment: ...

    def fetch(
        self,
        gps_start: int,
        gps_stop: int,
        channel_names: list[str],
    ) -> list[buffer]: ...

    def find_channels(
        self,
        channel_glob: str = ...,
        channel_type_mask: int = ...,
        data_type_mask: int = ...,
        min_sample_rate: float = ...,
        max_sample_rate: float = ...,
    ) -> list[channel]: ...

    def get_availability(
        self,
        channels: list[str],
    ) -> availability_list_type:
        ...

    def get_host(self) -> str: ...

    def get_port(self) -> int: ...

    def get_parameter(self, parameter: str) -> str: ...

    def get_protocol(self) -> int: ...

    @overload
    def iterate(
        self,
        channel_names: list[str],
    ) -> Iterator[Iterator[buffer]]: ...
    @overload
    def iterate(
        self,
        stride: int,
        channel_names: list[str],
    ) -> Iterator[Iterator[buffer]]: ...
    @overload
    def iterate(
        self,
        gps_start: int,
        gps_stop: int,
        channel_names: list[str],
    ) -> Iterator[Iterator[buffer]]: ...
    @overload
    def iterate(
        self,
        gps_start: int,
        gps_stop: int,
        stride: int,
        channel_names: list[str],
    ) -> Iterator[Iterator[buffer]]: ...

    @overload
    def set_epoch(self, start_time: str) -> None: ...
    @overload
    def set_epoch(self, start_time: int, end_time: int) -> None: ...

    def set_parameter(self, parameter: str, value: str) -> bool: ...

class segment:  # noqa: N801
    gps_start: int
    gps_stop: int
    frame_type: str

    @overload
    def __init__(self, start: int, stop: int) -> None: ...
    @overload
    def __init__(self, frame_type: str, start: int, stop: int) -> None: ...

class segment_list_type(list[segment]): ...  # noqa: N801

class simple_segment:  # noqa: N801
    gps_start: int
    gps_stop: int
